from __future__ import annotations

import dataclasses
from typing import Callable

from xarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.registry import Registry


def _reach_err(_, __):
    # can't raise error in lambda
    raise InvalidIcclimArgumentError(
        "Reach operator can't be called. Try to fill threshold with an operand"
        " (e.g. '>=' in '>= 22 degC')."
    )


@dataclasses.dataclass
class Operator:
    short_name: str
    long_name: str
    standard_name: str
    operand: str
    compute: Callable[[DataArray, DataArray | int | float], DataArray]
    aliases: list[str]

    def __call__(self, *args, **kwargs):
        return self.compute(*args, **kwargs)


class OperatorRegistry(Registry[Operator]):
    _item_class = Operator

    @staticmethod
    def get_item_aliases(op) -> list[str]:
        return list(map(str.upper, op.aliases))

    GREATER = Operator(
        short_name="gt",
        long_name="greater than",
        standard_name="greater_than",
        aliases=["gt", ">"],
        operand=">",
        compute=lambda da, th: da > th,  # noqa
    )
    LOWER = Operator(
        short_name="lt",
        long_name="lower than",
        standard_name="lower_than",
        aliases=["lt", "<"],
        operand="<",
        compute=lambda da, th: da < th,  # noqa
    )
    GREATER_OR_EQUAL = Operator(
        short_name="get",
        long_name="greater or equal to",
        standard_name="greater_or_equal_to",
        aliases=["get", "ge", ">=", "=>"],
        operand=">=",
        compute=lambda da, th: da >= th,  # noqa
    )
    LOWER_OR_EQUAL = Operator(
        short_name="let",
        long_name="lower or equal to",
        standard_name="lower_or_equal_to",
        aliases=["let", "le", "<=", "=<"],
        operand="<=",
        compute=lambda da, th: da <= th,  # noqa
    )
    EQUAL = Operator(
        short_name="e",
        long_name="equal to",
        standard_name="equal_to",
        aliases=["e", "equal", "eq", "=", "=="],
        operand="==",
        compute=lambda da, th: da == th,  # noqa
    )
    # A None operand means the threshold is reached and a reducer specific computation
    # is done. Case of excess and deficit (a.k.a gd4, hd17)
    REACH = Operator(
        short_name="reach",
        long_name="",  # nothing
        standard_name="reaching",
        aliases=["r"],
        operand="reach",
        compute=_reach_err,
    )



from __future__ import annotations

import abc
from abc import ABC
from datetime import timedelta
from functools import partial, reduce
from typing import Any, Callable
from warnings import warn

import jinja2
import numpy
import numpy as np
import xarray as xr
from jinja2 import Environment
from pint import DefinitionSyntaxError, Quantity, UndefinedUnitError
from xarray import DataArray
from xarray.core.resample import DataArrayResample
from xarray.core.rolling import DataArrayRolling
from xclim.core.calendar import select_time
from xclim.core.cfchecks import cfcheck_from_name
from xclim.core.datachecks import check_freq
from xclim.core.missing import MissingBase
from xclim.core.options import MISSING_METHODS, MISSING_OPTIONS, OPTIONS
from xclim.core.units import convert_units_to, rate2amount, str2pint, to_agg_units
from xclim.core.units import units as xc_units
from xclim.core.units import units2pint
from xclim.indices import run_length

from icclim.generic_indices.generic_templates import INDICATORS_TEMPLATES_EN
from icclim.generic_indices.threshold import PercentileThreshold, Threshold
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.cf_calendar import CfCalendarRegistry
from icclim.models.climate_variable import ClimateVariable
from icclim.models.constants import (
    GROUP_BY_METHOD,
    GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD,
    PART_OF_A_WHOLE_UNIT,
    REFERENCE_PERIOD_ID,
    RESAMPLE_METHOD,
    UNITS_KEY,
)
from icclim.models.frequency import RUN_INDEXER, Frequency, FrequencyRegistry
from icclim.models.index_config import IndexConfig
from icclim.models.logical_link import LogicalLink
from icclim.models.operator import OperatorRegistry
from icclim.models.registry import Registry

jinja_env = Environment()


class MissingMethodLike(metaclass=abc.ABCMeta):
    """workaround xclim missing type"""

    # todo: PR that to xclim

    def execute(self, *args, **kwargs) -> MissingBase:
        ...

    def validate(self, *args, **kwargs) -> bool:
        ...


class Indicator(ABC):
    name: str
    standard_name: str
    long_name: str
    cell_methods: str

    templated_properties = [
        "standard_name",
        "long_name",
        "cell_methods",
    ]

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> DataArray:
        ...

    @abc.abstractmethod
    def preprocess(self, *args, **kwargs) -> list[DataArray]:
        ...

    @abc.abstractmethod
    def postprocess(self, *args, **kwargs) -> DataArray:
        ...


class ResamplingIndicator(Indicator, ABC):
    """Abstract class for indicators."""

    missing: str
    missing_options: dict | None

    def __init__(self, missing="from_context", missing_options=None):
        self.missing_options = missing_options
        self.missing = missing
        if self.missing == "from_context" and self.missing_options is not None:
            raise ValueError(
                "Cannot set `missing_options` with `missing` method being from context."
            )
        missing_method: MissingMethodLike = MISSING_METHODS[self.missing]  # noqa typing
        self._missing = missing_method.execute
        if self.missing_options:
            missing_method.validate(**self.missing_options)
        super().__init__()

    def preprocess(
        self,
        climate_vars: list[ClimateVariable],
        jinja_scope: dict[str, Any],
        src_freq: Frequency,
    ) -> list[ClimateVariable]:
        _check_data(climate_vars, src_freq.pandas_freq)
        _check_cf(climate_vars)
        self.format(jinja_scope=jinja_scope)
        return climate_vars

    def postprocess(
        self,
        result: DataArray,
        climate_vars: list[ClimateVariable],
        output_freq: str,
        src_freq: str,
        indexer: dict,
        out_unit: str | None,
    ):
        if out_unit is not None:
            result = convert_units_to(result, out_unit)
        if self.missing != "skip" and indexer is not None:
            # reference variable is a subset of the studied variable,
            # so no need to check it.
            das = filter(lambda cv: not cv.is_reference, climate_vars)
            das = map(lambda cv: cv.studied_data, das)
            das = list(das)
            if "time" in result.dims:
                result = self._handle_missing_values(
                    resample_freq=output_freq,
                    src_freq=src_freq,
                    indexer=indexer,
                    in_data=das,
                    out_data=result,
                )
        for prop in self.templated_properties:
            result.attrs[prop] = getattr(self, prop)
        result.attrs["history"] = ""
        return result

    def format(self, jinja_scope: dict):
        for templated_property in self.templated_properties:
            template = jinja_env.from_string(
                getattr(self, templated_property),
                globals=jinja_scope,
            )
            setattr(self, templated_property, template.render())

    def _handle_missing_values(
        self,
        in_data: list[DataArray],
        resample_freq: str,
        src_freq: str,
        indexer: dict | None,
        out_data: DataArray,
    ) -> DataArray:
        options = self.missing_options or OPTIONS[MISSING_OPTIONS].get(self.missing, {})
        # We flag periods according to the missing method. skip variables without a time
        # coordinate.
        missing_method: MissingMethodLike = MISSING_METHODS[self.missing]  # noqa typing
        miss = (
            missing_method.execute(da, resample_freq, src_freq, options, indexer)
            for da in in_data
            if "time" in da.coords
        )
        # Reduce by or and broadcast to ensure the same length in time
        # When indexing is used and there are no valid points in the last period,
        # mask will not include it
        mask = reduce(np.logical_or, miss)  # noqa typing
        if isinstance(mask, DataArray) and mask.time.size < out_data.time.size:
            mask = mask.reindex(time=out_data.time, fill_value=True)
        return out_data.where(~mask)


class GenericIndicator(ResamplingIndicator):
    def __init__(
        self,
        name: str,
        process: Callable[..., DataArray],
        definition: str,
        check_vars: (
            Callable[[list[ClimateVariable], GenericIndicator], None] | None
        ) = None,
        sampling_methods: list[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        local = INDICATORS_TEMPLATES_EN
        self.name = name
        self.process = process
        self.standard_name = local[name]["standard_name"]
        self.cell_methods = local[name]["cell_methods"]
        self.long_name = local[name]["long_name"]
        self.check_vars = check_vars
        self.definition = definition
        self.sampling_methods = (
            sampling_methods if sampling_methods is not None else [RESAMPLE_METHOD]
        )

    def preprocess(  # noqa signature != from super
        self,
        climate_vars: list[ClimateVariable],
        jinja_scope: dict[str, Any],
        output_frequency: Frequency,
        src_freq: Frequency,
        output_unit: str | None,
        coef: float | None,
        sampling_method: str,
        is_compared_to_reference: bool,
    ) -> list[ClimateVariable]:
        if not _same_freq_for_all(climate_vars):
            raise InvalidIcclimArgumentError(
                "All variables must have the same time frequency (for example daily) to"
                " be compared with each others, but this was not the case."
            )
        if self.check_vars is not None:
            self.check_vars(climate_vars, self)
        if sampling_method not in self.sampling_methods:
            raise InvalidIcclimArgumentError(
                f"{self.name} can only be computed with the following"
                f" sampling_method(s): {self.sampling_methods}"
            )
        if output_unit is not None and _is_amount_unit(output_unit):
            for climate_var in climate_vars:
                current_unit = climate_var.studied_data.attrs.get(UNITS_KEY, None)
                if current_unit is not None and not _is_amount_unit(current_unit):
                    climate_var.studied_data = rate2amount(
                        climate_var.studied_data, out_units=output_unit
                    )
        if coef is not None:
            for climate_var in climate_vars:
                climate_var.studied_data = coef * climate_var.studied_data
        if output_frequency.indexer:
            for climate_var in climate_vars:
                climate_var.studied_data = select_time(
                    climate_var.studied_data, **output_frequency.indexer, drop=True
                )
        return super().preprocess(
            climate_vars=climate_vars,
            jinja_scope=jinja_scope,
            src_freq=src_freq,
        )

    def __call__(self, config: IndexConfig) -> DataArray:
        src_freq = config.climate_variables[0].source_frequency
        base_jinja_scope = {
            "np": numpy,
            "enumerate": enumerate,
            "len": len,
            "output_freq": config.frequency,
            "source_freq": src_freq,
        }
        climate_vars_meta = _get_climate_vars_metadata(
            config.climate_variables, src_freq, base_jinja_scope, jinja_env
        )
        jinja_scope: dict[str, Any] = {
            "min_spell_length": config.min_spell_length,
            "rolling_window_width": config.rolling_window_width,
            "climate_vars": climate_vars_meta,
            "is_compared_to_reference": config.is_compared_to_reference,
            "reference_period": config.reference_period,
        }
        jinja_scope.update(base_jinja_scope)
        climate_vars = self.preprocess(
            climate_vars=config.climate_variables,
            jinja_scope=jinja_scope,
            output_frequency=config.frequency,
            src_freq=src_freq,
            output_unit=config.out_unit,
            coef=config.coef,
            sampling_method=config.sampling_method,
            is_compared_to_reference=config.is_compared_to_reference,
        )
        result = self.process(
            climate_vars=climate_vars,
            resample_freq=config.frequency,
            min_spell_length=config.min_spell_length,
            rolling_window_width=config.rolling_window_width,
            group_by_freq=config.frequency.group_by_key,
            is_compared_to_reference=config.is_compared_to_reference,
            logical_link=config.logical_link,
            date_event=config.date_event,
            source_freq_delta=src_freq.delta,
            to_percent=config.out_unit == "%",
            sampling_method=config.sampling_method,
        )
        return self.postprocess(
            result,
            climate_vars=climate_vars,
            output_freq=config.frequency.pandas_freq,
            src_freq=src_freq.pandas_freq,
            indexer=config.frequency.indexer,
            out_unit=config.out_unit,
        )

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, GenericIndicator)
            and self.long_name == other.long_name
            and self.standard_name == other.standard_name
            and self.process == other.process
        )


def count_occurrences(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    date_event: bool,
    to_percent: bool,
    **kwargs,  # noqa
) -> DataArray:
    if date_event:
        reducer_op = _count_occurrences_with_date
    else:
        reducer_op = partial(DataArray.sum, dim="time")
    merged_exceedances = _compute_exceedances(
        climate_vars, resample_freq.pandas_freq, logical_link
    )
    result = reducer_op(merged_exceedances.resample(time=resample_freq.pandas_freq))
    if to_percent:
        result = _to_percent(result, resample_freq)
        result.attrs[UNITS_KEY] = "%"
        return result
    else:
        return to_agg_units(result, climate_vars[0].studied_data, "count")


def max_consecutive_occurrence(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa
) -> DataArray:
    merged_exceedances = _compute_exceedances(
        climate_vars, resample_freq.pandas_freq, logical_link
    )
    rle = run_length.rle(merged_exceedances, dim="time", index="first")
    resampled = rle.resample(time=resample_freq.pandas_freq)
    if date_event:
        result = _consecutive_occurrences_with_dates(resampled, source_freq_delta)
    else:
        result = resampled.max(dim="time")
    return to_agg_units(result, climate_vars[0].studied_data, "count")


def sum_of_spell_lengths(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    min_spell_length: int,
    **kwargs,  # noqa
) -> DataArray:
    merged_exceedances = _compute_exceedances(
        climate_vars, resample_freq.pandas_freq, logical_link
    )
    rle = run_length.rle(merged_exceedances, dim="time", index="first")
    cropped_rle = rle.where(rle >= min_spell_length, other=0)
    result = cropped_rle.resample(time=resample_freq.pandas_freq).max(dim="time")
    return to_agg_units(result, climate_vars[0].studied_data, "count")


def excess(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    study, threshold = get_single_var(climate_vars)
    if threshold.operator is not OperatorRegistry.REACH:
        raise InvalidIcclimArgumentError("")
    excesses = threshold.compute(study, override_op=lambda da, th: da - th)
    res = (
        (excesses).clip(min=0).resample(time=resample_freq.pandas_freq).sum(dim="time")
    )
    return to_agg_units(res, study, "delta_prod")


def deficit(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    study, threshold = get_single_var(climate_vars)
    deficit = threshold.compute(study, override_op=lambda da, th: th - da)
    res = deficit.clip(min=0).resample(time=resample_freq.pandas_freq).sum(dim="time")
    return to_agg_units(res, study, "delta_prod")


def fraction_of_total(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    to_percent: bool,
    **kwargs,  # noqa
) -> DataArray:
    study, threshold = get_single_var(climate_vars)
    if threshold.threshold_min_value is not None:
        total = (
            study.where(threshold.operator(study, threshold.threshold_min_value.m))
            .resample(time=resample_freq.pandas_freq)
            .sum(dim="time")
        )
    else:
        total = study.resample(time=resample_freq.pandas_freq).sum(dim="time")
    exceedance = _compute_exceedance(
        study=study,
        threshold=threshold,
        freq=resample_freq.pandas_freq,
        bootstrap=_must_run_bootstrap(study, threshold),
    ).squeeze()
    over = (
        study.where(exceedance, 0)
        .resample(time=resample_freq.pandas_freq)
        .sum(dim="time")
    )
    res = over / total
    if to_percent:
        res = res * 100
        res.attrs[UNITS_KEY] = "%"
    else:
        res.attrs[UNITS_KEY] = PART_OF_A_WHOLE_UNIT
    return res


def maximum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    date_event: bool,
    **kwargs,  # noqa
) -> DataArray:
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.max,
        date_event=date_event,
    )


def minimum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    date_event: bool,
    **kwargs,  # noqa
) -> DataArray:
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.min,
        date_event=date_event,
    )


def average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.mean,
        date_event=False,
    )


def sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.sum,
        date_event=False,
        must_convert_rate=True,
    )


def standard_deviation(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    return _run_simple_reducer(
        climate_vars, resample_freq, DataArrayResample.std, date_event=False
    )


def max_of_rolling_sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa
):
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.sum,
        resampled_op=DataArrayResample.max,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def min_of_rolling_sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa
):
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.sum,
        resampled_op=DataArrayResample.min,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def min_of_rolling_average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa
):
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.mean,
        resampled_op=DataArrayResample.min,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def max_of_rolling_average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa
):
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.mean,
        resampled_op=DataArrayResample.max,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def mean_of_difference(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
):
    study, ref = get_couple_of_var(climate_vars, "mean_of_difference")
    mean_of_diff = (study - ref).resample(time=resample_freq.pandas_freq).mean()
    mean_of_diff.attrs["units"] = study.attrs["units"]
    return mean_of_diff


def difference_of_extremes(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
):
    study, ref = get_couple_of_var(climate_vars, "difference_of_extremes")
    max_study = study.resample(time=resample_freq.pandas_freq).max()
    min_ref = ref.resample(time=resample_freq.pandas_freq).min()
    diff_of_extremes = max_study - min_ref
    diff_of_extremes.attrs["units"] = study.attrs["units"]
    return diff_of_extremes


def mean_of_absolute_one_time_step_difference(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa
) -> DataArray:
    """
    Generification of ECAD's vDTR index.

    Parameters
    ----------
    climate_vars : List[ClimateVariable]
    The two climate variables necessary to compute the indicator.
    resample_freq :  Frequency
    Expected frequency of the output.
    kwargs : dict
    Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
    mean_of_absolute_one_time_step_difference as a xarray.DataArray
    """
    study, ref = get_couple_of_var(
        climate_vars, "mean_of_absolute_one_time_step_difference"
    )
    one_time_step_diff = (study - ref).diff(dim="time")
    res = abs(one_time_step_diff).resample(time=resample_freq.pandas_freq).mean()
    res.attrs["units"] = study.attrs["units"]
    return res


def difference_of_means(
    climate_vars: list[ClimateVariable],
    to_percent: bool,
    resample_freq: Frequency,
    sampling_method: str,
    is_compared_to_reference: bool,
    **kwargs,  # noqa
):
    if is_compared_to_reference and sampling_method == RESAMPLE_METHOD:
        raise InvalidIcclimArgumentError(
            "It does not make sense to resample the reference variable if it is"
            " already a subsample of the studied variable. Try setting"
            f" `sampling_method='{GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD}'`"
            f" instead."
        )
    study, ref = get_couple_of_var(climate_vars, "difference_of_means")
    if sampling_method == GROUP_BY_METHOD:
        if resample_freq.group_by_key == RUN_INDEXER:
            mean_study = study.mean(dim="time")
            mean_ref = ref.mean(dim="time")
        else:
            mean_study = study.groupby(resample_freq.group_by_key).mean()
            mean_ref = ref.groupby(resample_freq.group_by_key).mean()
    elif sampling_method == RESAMPLE_METHOD:
        mean_study = study.resample(time=resample_freq.pandas_freq).mean()
        mean_ref = ref.resample(time=resample_freq.pandas_freq).mean()
    elif sampling_method == GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD:
        if (
            resample_freq.group_by_key == RUN_INDEXER
            or resample_freq == FrequencyRegistry.YEAR
        ):
            mean_study = study.resample(time=resample_freq.pandas_freq).mean()
            # data is already filtered with only the indexed values.
            # Thus there is only one "group".
            mean_ref = ref.mean(dim="time")
        else:
            return _diff_of_means_of_resampled_x_by_groupedby_y(
                resample_freq, to_percent, study, ref
            )
    else:
        raise NotImplementedError(f"Unknown sampling_method: '{sampling_method}'.")
    diff_of_means = mean_study - mean_ref
    if to_percent:
        diff_of_means = diff_of_means / mean_ref * 100
        diff_of_means.attrs["units"] = "%"
    else:
        diff_of_means.attrs["units"] = study.attrs["units"]
    return diff_of_means


def _diff_of_means_of_resampled_x_by_groupedby_y(
    resample_freq: Frequency, to_percent: bool, study: DataArray, ref: DataArray
) -> DataArray:
    mean_ref = ref.groupby(resample_freq.group_by_key).mean()
    acc = []
    if resample_freq == FrequencyRegistry.MONTH:
        key = "month"
        dt_selector = lambda x: x.time.dt.month  # noqa lamdab assigned
    elif resample_freq == FrequencyRegistry.DAY:
        key = "dayofyear"
        dt_selector = lambda x: x.time.dt.dayofyear  # noqa lamdab assigned
    else:
        raise NotImplementedError(
            f"Can't use {GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD}"
            f" with the frequency {resample_freq.long_name}."
        )
    for label, sample in study.resample(time=resample_freq.pandas_freq):
        sample_mean = sample.mean(dim="time")
        ref_group_mean = mean_ref.sel({key: dt_selector(sample).values[0]})
        sample_diff_of_means = sample_mean - ref_group_mean
        if to_percent:
            sample_diff_of_means = sample_diff_of_means / ref_group_mean * 100
        del sample_diff_of_means[key]
        sample_diff_of_means = sample_diff_of_means.expand_dims(time=[label])
        acc.append(sample_diff_of_means)
    diff_of_means = xr.concat(acc, dim="time")
    if to_percent:
        diff_of_means.attrs["units"] = "%"
    else:
        diff_of_means.attrs["units"] = study.attrs["units"]
    return diff_of_means


def _check_single_var(climate_vars: list[ClimateVariable], indicator: GenericIndicator):
    if len(climate_vars) > 1:
        raise InvalidIcclimArgumentError(
            f"{indicator.name} can only be computed on a single variable."
        )


def _check_couple_of_vars(
    climate_vars: list[ClimateVariable], indicator: GenericIndicator
):
    if len(climate_vars) != 2:
        raise InvalidIcclimArgumentError(
            f"{indicator.name} can only be computed on two variables sharing the same"
            f" unit (e.g. 2 temperatures). Either provide a `base_period_time_range` to"
            f" create a reference variable or directly provide a secondary variable"
            f" with `in_files` or `var_name`."
        )


class GenericIndicatorRegistry(Registry[GenericIndicator]):
    def __init__(self):
        super().__init__()

    _item_class = GenericIndicator

    CountOccurrences = GenericIndicator(
        "count_occurrences",
        count_occurrences,
        definition="Count occurrences where threshold(s) are met"
        " (e.g. SU, Tx90p, RR1).",
    )
    MaxConsecutiveOccurrence = GenericIndicator(
        "max_consecutive_occurrence",
        max_consecutive_occurrence,
        definition="Count the maximum number of consecutive occurrences when"
        " threshold(s) are met (e.g. CDD, CSU, CWD).",
    )
    SumOfSpellLengths = GenericIndicator(
        "sum_of_spell_lengths",
        sum_of_spell_lengths,
        definition="Sum the lengths of each consecutive occurrence spell when"
        " threshold(s) are met. The minimum spell length is controlled by"
        " `min_spell_length` (e.g. WSDI, CSDI).",
    )
    Excess = GenericIndicator(
        "excess",
        excess,
        check_vars=_check_single_var,
        definition="Compute the excess over the given threshold. The excess is"
        " `sum(x[x>t] - t)` where x is the studied variable and t the threshold"
        " (e.g. GD4).",
    )
    Deficit = GenericIndicator(
        "deficit",
        deficit,
        check_vars=_check_single_var,
        definition="Compute the deficit below the given threshold. The deficit is"
        " `sum(t - x[x<t])` where x is the studied variable and t the threshold"
        " (e.g. HD17).",
    )
    FractionOfTotal = GenericIndicator(
        "fraction_of_total",
        fraction_of_total,
        check_vars=_check_single_var,
        definition="Compute the fraction of values meeting threshold(s) over the sum of"
        " every values (e.g. R75pTOT, R95pTOT).",
    )
    Maximum = GenericIndicator(
        "maximum",
        maximum,
        definition="Maximum of values that met threshold(s), if threshold(s) are given"
        " (e.g. Txx, Tnx).",
    )
    Minimum = GenericIndicator(
        "minimum",
        minimum,
        definition="Maximum of values that met threshold(s), if threshold(s) are given"
        " (e.g. Txn, Tnn).",
    )
    Average = GenericIndicator(
        "average",
        average,
        definition="Average of values that met threshold(s), if threshold(s) are given"
        " (e.g. Tx, Tn)",
    )
    Sum = GenericIndicator(
        "sum",
        sum,
        definition="Sum of values that met threshold(s), if threshold(s) are given"
        " (e.g. PRCPTOT, RR).",
    )
    StandardDeviation = GenericIndicator(
        "standard_deviation",
        standard_deviation,
        definition="Standard deviation of values that met threshold(s),"
        " if threshold(s) are given.",
    )
    MaxOfRollingSum = GenericIndicator(
        "max_of_rolling_sum",
        max_of_rolling_sum,
        check_vars=_check_single_var,
        definition="Maximum of rolling sum over time dimension"
        " (e.g. RX5DAY: maximum 5 days window of precipitation accumulation).",
    )
    MinOfRollingSum = GenericIndicator(
        "min_of_rolling_sum",
        min_of_rolling_sum,
        check_vars=_check_single_var,
        definition="Minimum of rolling sum over time dimension.",
    )
    MaxOfRollingAverage = GenericIndicator(
        "max_of_rolling_average",
        max_of_rolling_average,
        check_vars=_check_single_var,
        definition="Maximum of rolling average over time dimension.",
    )
    MinOfRollingAverage = GenericIndicator(
        "min_of_rolling_average",
        min_of_rolling_average,
        check_vars=_check_single_var,
        definition="Minimum of rolling average over time dimension.",
    )
    MeanOfDifference = GenericIndicator(
        "mean_of_difference",
        mean_of_difference,
        check_vars=_check_couple_of_vars,
        definition="Average of the difference between two variables"
        ", or one variable and it's reference period values"
        " (e.g. DTR: `mean(tasmax - tasmin)`).",
    )
    DifferenceOfExtremes = GenericIndicator(
        "difference_of_extremes",
        difference_of_extremes,
        check_vars=_check_couple_of_vars,
        definition="Difference of extremes between two variables"
        ", or one variable and it's reference period values."
        " The extremes are always `maximum` for the first variable and"
        " `minimum` for the second variable"
        " (e.g. ETR: `max(tasmax) - min(tasmin)`).",
    )
    MeanOfAbsoluteOneTimeStepDifference = GenericIndicator(
        "mean_of_absolute_one_time_step_difference",
        mean_of_absolute_one_time_step_difference,
        check_vars=_check_couple_of_vars,
        definition="Average of the absolute one time step by one time step difference"
        " between two variables,"
        " or one variable and it's reference period values"
        " (e.g. vDTR:"
        " `mean((tasmax[i] - tasmin[i]) - (tasmax[i-1] - tasmin[i-1])` ;"
        " where i is the day of measure).",
    )
    DifferenceOfMeans = GenericIndicator(
        "difference_of_means",
        difference_of_means,
        check_vars=_check_couple_of_vars,
        sampling_methods=[
            RESAMPLE_METHOD,
            GROUP_BY_METHOD,
            GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD,
        ],
        definition="Difference of the average between two variables"
        ", or one variable and it's reference period values"
        " (e.g. anomaly: `mean(tasmax) - mean(tasmax_ref]))`.",
    )


def _compute_exceedance(
    study: DataArray,
    threshold: Threshold,
    freq: str,  # noqa used by @percentile_bootstrap (don't rename, it breaks bootstrap)
    bootstrap: bool,  # noqa used by @percentile_bootstrap
) -> DataArray:
    exceedances = threshold.compute(study, freq=freq, bootstrap=bootstrap)
    if bootstrap:
        exceedances.attrs[REFERENCE_PERIOD_ID] = threshold.value.attrs[
            "climatology_bounds"
        ]
    return exceedances


def get_couple_of_var(
    climate_vars: list[ClimateVariable], indicator: str
) -> tuple[DataArray, DataArray]:
    if len(climate_vars) != 2:
        raise InvalidIcclimArgumentError(
            f"{indicator} needs two variables **or** one variable and a "
            f"`base_period_time_range` period to extract a reference variable."
        )
    if climate_vars[0].threshold or climate_vars[1].threshold:
        raise InvalidIcclimArgumentError(
            f"{indicator} cannot be computed with thresholds."
        )
    study = climate_vars[0].studied_data
    ref = climate_vars[1].studied_data
    study = convert_units_to(study, ref)
    return study, ref


def _run_rolling_reducer(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    rolling_op: Callable[[DataArrayRolling], DataArray],  # sum | mean
    resampled_op: Callable[[...], DataArray],  # max | min
    date_event: bool,
    source_freq_delta: timedelta,
) -> DataArray:
    study, threshold = get_single_var(climate_vars)
    if threshold:
        exceedance = _compute_exceedance(
            study=study,
            freq=resample_freq.pandas_freq,
            threshold=threshold,
            bootstrap=_must_run_bootstrap(study, threshold),
        ).squeeze()
        study = study.where(exceedance)
    study = rolling_op(study.rolling(time=rolling_window_width))
    study = study.resample(time=resample_freq.pandas_freq)
    if date_event:
        return _reduce_with_date_event(
            resampled=study,
            reducer=resampled_op,
            window=rolling_window_width,
            source_delta=source_freq_delta,
        )
    else:
        return resampled_op(study, dim="time")  # type:ignore


def _run_simple_reducer(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    reducer_op: Callable[..., DataArray],
    date_event: bool,
    must_convert_rate: bool = False,
):
    study, threshold = get_single_var(climate_vars)
    if threshold is not None:
        exceedance = _compute_exceedance(
            study=study,
            freq=resample_freq.pandas_freq,
            threshold=threshold,
            bootstrap=_must_run_bootstrap(study, threshold),
        ).squeeze()
        filtered_study = study.where(exceedance)
    else:
        filtered_study = study
    if must_convert_rate:
        if _is_rate(filtered_study):
            filtered_study = rate2amount(filtered_study)
    if date_event:
        return _reduce_with_date_event(
            resampled=filtered_study.resample(time=resample_freq.pandas_freq),
            reducer=reducer_op,
        )
    else:
        return reducer_op(
            filtered_study.resample(time=resample_freq.pandas_freq), dim="time"
        )


def _compute_exceedances(
    climate_vars: list[ClimateVariable], resample_freq: str, logical_link: LogicalLink
) -> DataArray:
    exceedances = [
        _compute_exceedance(
            study=climate_var.studied_data,
            threshold=climate_var.threshold,
            freq=resample_freq,
            bootstrap=_must_run_bootstrap(
                climate_var.studied_data, climate_var.threshold
            ),
        ).squeeze()
        for climate_var in climate_vars
    ]
    return logical_link(exceedances)


def get_single_var(
    climate_vars: list[ClimateVariable],
) -> tuple[DataArray, Threshold | None]:
    if climate_vars[0].threshold:
        return (
            climate_vars[0].studied_data,
            climate_vars[0].threshold,
        )
    else:
        return climate_vars[0].studied_data, None


def _must_run_bootstrap(da: DataArray, threshold: Threshold | None) -> bool:
    """Avoid bootstrapping if there is one single year overlapping
    or no year overlapping or all year overlapping.
    """
    # TODO: Don't run bootstrap when not on extreme percentile
    #       (run only below 20? 10? and above 80? 90?)
    if (
        threshold is None
        or not isinstance(threshold, PercentileThreshold)
        or (
            isinstance(threshold, PercentileThreshold)
            and not threshold.is_doy_per_threshold
        )
    ):
        return False
    reference = threshold.value
    study_years = np.unique(da.indexes.get("time").year)
    overlapping_years = np.unique(
        da.sel(time=_get_ref_period_slice(reference)).indexes.get("time").year
    )
    return 1 < len(overlapping_years) < len(study_years)


def _get_ref_period_slice(da: DataArray) -> slice:
    if (bds := da.attrs.get("climatology_bounds", None)) is not None:
        return slice(*bds)
    time_length = len(da.time)
    return slice(*da.time[0 :: time_length - 1].dt.strftime("%Y-%m-%d").values)


def _same_freq_for_all(climate_vars: list[ClimateVariable]) -> bool:
    if len(climate_vars) == 1:
        return True
    freqs = list(map(lambda a: xr.infer_freq(a.studied_data.time), climate_vars))
    return all(map(lambda x: x == freqs[0], freqs[1:]))


def _get_climate_vars_metadata(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    jinja_scope: dict[str, Any],
    jinja_env: jinja2.Environment,
) -> list[dict[str, str]]:
    return [
        c_var.build_indicator_metadata(
            resample_freq,
            _must_run_bootstrap(c_var.studied_data, c_var.threshold),
            jinja_scope,
            jinja_env,
        )
        for c_var in climate_vars
    ]


def _reduce_with_date_event(
    resampled: DataArrayResample,
    reducer: Callable[[DataArrayResample], DataArray],
    source_delta: timedelta | None = None,
    window: int | None = None,
) -> DataArray:
    acc: list[DataArray] = []
    if reducer == DataArrayResample.max:
        group_reducer = DataArray.argmax
    elif reducer == DataArrayResample.min:
        group_reducer = DataArray.argmin
    else:
        raise NotImplementedError(
            f"Can't compute `date_event` due to unknown reducer:" f" '{reducer}'"
        )
    for label, sample in resampled:
        reduced_result = sample.isel(time=group_reducer(sample, dim="time"))
        if window is not None:
            result = _add_date_coords(
                original_sample=sample,
                result=sample.sum(dim="time"),
                start_time=reduced_result.time,
                end_time=reduced_result.time + window * source_delta,
                label=label,
            )
        else:
            result = _add_date_coords(
                original_sample=sample,
                result=sample.sum(dim="time"),
                event_date=reduced_result.time,
                label=label,
            )
        acc.append(result)
    return xr.concat(acc, "time")


def _count_occurrences_with_date(resampled: DataArrayResample):
    acc: list[DataArray] = []
    for label, sample in resampled:
        # Fixme probably not safe to compute on huge dataset,
        #  it should be fixed with
        #  https://github.com/pydata/xarray/issues/2511
        sample = sample.compute()
        first = sample.isel(time=sample.argmax("time")).time
        reversed_time = sample.reindex(time=list(reversed(sample.time.values)))
        last = reversed_time.isel(time=reversed_time.argmax("time")).time
        dated_occurrences = _add_date_coords(
            original_sample=sample,
            result=sample.sum(dim="time"),
            start_time=first,
            end_time=last,
            label=label,
        )
        acc.append(dated_occurrences)
    return xr.concat(acc, "time")


def _consecutive_occurrences_with_dates(
    resampled: DataArrayResample, source_freq_delta: timedelta
):
    acc = []
    for label, sample in resampled:
        sample = sample.where(~sample.isnull(), 0)
        time_index_of_max_rle = sample.argmax(dim="time")
        # fixme: `.compute` is needed until xarray merges this pr:
        #        https://github.com/pydata/xarray/pull/5873
        time_index_of_max_rle = time_index_of_max_rle.compute()
        dated_longest_run = sample[{"time": time_index_of_max_rle}]
        start_time = sample.isel(
            time=time_index_of_max_rle.where(time_index_of_max_rle > 0, 0)
        ).time
        end_time = start_time + (dated_longest_run * source_freq_delta)
        dated_longest_run = _add_date_coords(
            original_sample=sample,
            result=dated_longest_run,
            start_time=start_time,
            end_time=end_time,
            label=label,
        )
        acc.append(dated_longest_run)
    result = xr.concat(acc, "time")
    return result


def _add_date_coords(
    original_sample: DataArray,
    result: DataArray,
    label: str | np.datetime64,
    start_time: DataArray = None,
    end_time: DataArray = None,
    event_date: DataArray = None,
) -> DataArray:
    new_coords = {c: original_sample[c] for c in original_sample.coords if c != "time"}
    if event_date is None:
        new_coords["event_date_start"] = start_time
        new_coords["event_date_end"] = end_time
    else:
        new_coords["event_date"] = event_date
    new_coords["time"] = label
    return DataArray(data=result, coords=new_coords)


def _is_amount_unit(unit: str) -> bool:
    try:
        u = units2pint(unit)  # turn a cf u
        return xc_units.Quantity(1, u).check("[length]")
    except (UndefinedUnitError, DefinitionSyntaxError):
        return False


def _to_percent(da: DataArray, sampling_freq: Frequency) -> DataArray:
    if sampling_freq == FrequencyRegistry.MONTH:
        da = da / da.time.dt.daysinmonth * 100
    elif sampling_freq == FrequencyRegistry.YEAR:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 366
        coef[{"time": ~leap_years}] = 365
        da = da / coef
    elif sampling_freq == FrequencyRegistry.AMJJAS:
        da = da / 183
    elif sampling_freq == FrequencyRegistry.ONDJFM:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 183
        coef[{"time": ~leap_years}] = 182
        da = da / coef
    elif sampling_freq == FrequencyRegistry.DJF:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 91
        coef[{"time": ~leap_years}] = 90
        da = da / coef
    elif sampling_freq in [FrequencyRegistry.MAM, FrequencyRegistry.JJA]:
        da = da / 92
    elif sampling_freq == FrequencyRegistry.SON:
        da = da / 91
    else:
        # TODO improve this for custom resampling
        warn(
            "For now, '%' unit can only be used when `slice_mode` is one of: "
            "{MONTH, YEAR, AMJJAS, ONDJFM, DJF, MAM, JJA, SON}."
        )
        return da
    da.attrs[UNITS_KEY] = PART_OF_A_WHOLE_UNIT
    return da


def _is_leap_year(da: DataArray) -> np.ndarray:
    time_index = da.indexes.get("time")
    if isinstance(time_index, xr.CFTimeIndex):
        return CfCalendarRegistry.lookup(time_index.calendar).is_leap(da.time.dt.year)
    else:
        return da.time.dt.is_leap_year


def _check_cf(climate_vars: list[ClimateVariable]):
    """Compare metadata attributes to CF-Convention standards.

    Default cfchecks use the specifications in `xclim.core.utils.VARIABLES`,
    assuming the indicator's inputs are using the CMIP6/xclim variable names
    correctly.
    Variables absent from these default specs are silently ignored.

    When subclassing this method, use functions decorated using
    `xclim.core.options.cfcheck`.
    """
    for da in climate_vars:
        try:
            cfcheck_from_name(str(da.name), da)
        except KeyError:
            # Silently ignore unknown variables.
            pass


def _check_data(climate_vars: list[ClimateVariable], src_freq: str):
    if src_freq is None:
        return
    for climate_var in climate_vars:
        da = climate_var.studied_data
        if "time" in da.coords and da.time.ndim == 1 and len(da.time) > 3:
            check_freq(da, src_freq, strict=True)


def _is_rate(query: Quantity | DataArray) -> bool:
    if isinstance(query, DataArray):
        query = str2pint(query.attrs[UNITS_KEY])
    return query.dimensionality.get("[time]", None) == -1



from __future__ import annotations

from xclim.core.calendar import build_climatology_bounds

import icclim
from icclim.models.constants import (
    UNITS_KEY,
    USER_INDEX_PRECIPITATION_STAMP,
    USER_INDEX_TEMPERATURE_STAMP,
)
from icclim.models.operator import OperatorRegistry
from icclim.tests.testing_utils import stub_tas


class Test_max:
    def test_simple(self):
        da = stub_tas(use_dask=False)
        da.data[1] = 20
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(
                index_name="data", calc_operation="max", coef=1, logical_operation=None
            ),
        )
        # THEN
        assert result.data[0] == 20


class Test_min:
    def test_simple(self):
        da = stub_tas(use_dask=False)
        da.data[1] = -20
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(index_name="data", calc_operation="min"),
        )
        # THEN
        assert result.data[0] == -20


class Test_mean:
    def test_simple(self):
        da = stub_tas(use_dask=False)
        da[2] = 366
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(index_name="data", calc_operation="mean"),
        )
        # THEN
        assert result.data[0] == 2


class Test_sum:
    def test_simple(self):
        da = stub_tas(use_dask=False)
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(index_name="data", calc_operation="sum"),
            slice_mode="year",
        )
        # THEN
        assert result.data[0] == 365


class Test_count_events:
    def test_simple(self):
        # GIVEN
        da = stub_tas(10, False)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(
                index_name="data",
                calc_operation="nb_events",
                thresh=15,
                logical_operation=OperatorRegistry.GREATER,
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 1

    def test_simple_default_percentile(self):
        # GIVEN
        da = stub_tas(10, False)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(
                index_name="data",
                calc_operation="nb_events",
                thresh="50p",
                logical_operation=OperatorRegistry.GREATER,
            ),
            base_period_time_range=build_climatology_bounds(da),
            slice_mode="month",
        )
        # THEN
        assert result.data.isel(time=0) == 2

    def test_simple_period_percentile(self):
        # GIVEN
        da = stub_tas(10, False)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(
                index_name="data",
                calc_operation="nb_events",
                thresh="50p",
                var_type=USER_INDEX_PRECIPITATION_STAMP,
                logical_operation=OperatorRegistry.GREATER,
            ),
            base_period_time_range=build_climatology_bounds(da),
            slice_mode="month",
        )
        # THEN
        assert result.data.isel(time=0) == 2

    def test_simple_doy_percentile(self):
        # GIVEN
        da = stub_tas(10, False)
        da[1] = 15
        da[2] = 16
        # WHEN
        result = icclim.index(
            in_files=da,
            user_index=dict(
                index_name="data",
                calc_operation="nb_events",
                thresh="80p",
                var_type=USER_INDEX_TEMPERATURE_STAMP,
                logical_operation=OperatorRegistry.GREATER,
            ),
            base_period_time_range=build_climatology_bounds(da),
            slice_mode="month",
        )
        # THEN
        assert result.data.isel(time=0) == 2

    def test_multi_threshold_or(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[1] = 15
        tmin = stub_tas(-10, False)
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax, "tmin": tmin},
            index_name="data",
            user_index=dict(
                calc_operation="nb_events",
                thresh=[12, -20],
                var_type=USER_INDEX_TEMPERATURE_STAMP,
                logical_operation=[OperatorRegistry.GREATER, OperatorRegistry.EQUAL],
                link_logical_operations="or",
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 1

    def test_multi_threshold_and(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[1] = 15
        tmin = stub_tas(-10, False)
        tmin[1] = -20
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax, "tmin": tmin},
            index_name="data",
            user_index=dict(
                calc_operation="nb_events",
                thresh=[12, -20],
                var_type=USER_INDEX_TEMPERATURE_STAMP,
                logical_operation=[OperatorRegistry.GREATER, OperatorRegistry.EQUAL],
                link_logical_operations="and",
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 1


class Test_run_mean:
    def test_run_mean_min(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[30] = 0
        tmax[29] = 0
        tmax[28] = 0
        tmax[27] = 0
        tmax[26] = 0
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            index_name="data",
            user_index=dict(
                calc_operation="run_mean",
                extreme_mode="min",
                window_width=5,
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 0
        assert result.data[1] == 2
        assert result.data[2] == 10

    def test_run_mean_max(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[30] = 20
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            index_name="data",
            rolling_window_width=2,
            user_index=dict(
                calc_operation="run_mean",
                extreme_mode="max",
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 15
        assert result.data[1] == 15
        assert result.data[2] == 10


class Test_run_sum:
    def test_run_sum_min(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[30] = 0
        tmax[29] = 0
        tmax[28] = 0
        tmax[27] = 0
        tmax[26] = 0
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            index_name="data",
            rolling_window_width=5,
            user_index=dict(
                calc_operation="run_sum",
                extreme_mode="min",
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 0
        assert result.data[1] == 10
        assert result.data[2] == 50

    def test_run_sum_max(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[30] = 20
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            index_name="data",
            rolling_window_width=2,
            user_index=dict(
                calc_operation="run_sum",
                extreme_mode="max",
            ),
            slice_mode="month",
        )
        # THEN
        assert result.data[0] == 30
        assert result.data[1] == 30
        assert result.data[2] == 20


class Test_max_consecutive_event_count:
    def test_simple(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax[30] = 15  # On 31th january
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            index_name="data",
            user_index=dict(
                calc_operation="max_nb_consecutive_events",
                thresh=10.0,
                logical_operation=OperatorRegistry.EQUAL,
            ),
            slice_mode="year",
        )
        # THEN
        assert result.data[0] == 1795
        assert result.data[1].isnull()


class Test_anomaly:
    def test_simple(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax2 = stub_tas(11, False)
        # WHEN
        result = icclim.index(
            in_files={"tmax2": tmax2, "tmax": tmax},
            index_name="data",
            user_index=dict(
                calc_operation="anomaly",
            ),
            slice_mode="year",
        )
        # THEN
        assert (result.data == 1).all()
        assert result.data.attrs[UNITS_KEY] == tmax.attrs[UNITS_KEY]

    def test_single_var(self):
        # GIVEN
        tmax = stub_tas(10, False)
        first_year = tmax.time.dt.year.min().values[()]
        tmax = tmax.where(tmax.time.dt.year <= first_year + 1, 11)  #
        ref = tmax.sel(time=slice(str(first_year), str(first_year + 1)))
        ref_bds = build_climatology_bounds(ref)
        study = tmax.where(~tmax.time.dt.year.isin(ref.time.dt.year), drop=True)
        study_bds = build_climatology_bounds(study)
        # WHEN
        result = icclim.index(
            in_files={"tmax": tmax},
            time_range=study_bds,
            base_period_time_range=ref_bds,
            index_name="data",
            sampling_method="groupby",
            user_index=dict(
                calc_operation="anomaly",
            ),
            slice_mode="month",
        )
        # THEN
        assert (result.data == 1).all()
        assert len(result.data.month) == 12
        assert result.data.attrs[UNITS_KEY] == tmax.attrs[UNITS_KEY]

    def test_simple_percent(self):
        # GIVEN
        tmax = stub_tas(10, False)
        tmax2 = stub_tas(11, False)
        # WHEN
        result = icclim.index(
            index_name="data",
            in_files={"tmax2": tmax2, "tmax": tmax},
            out_unit="%",
            user_index=dict(calc_operation="anomaly"),
        )
        # THEN
        assert (result.data == 10).all()
        assert result.data.attrs[UNITS_KEY] == "%"



"""
    `icclim.models.frequency` wraps the concept of pandas frequency in order to resample
    time series. `slice_mode` parameter of `icclim.index` is always converted to a
    `Frequency`.
"""
from __future__ import annotations

import dataclasses
import re
from datetime import timedelta
from typing import Any, Callable, Sequence

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from pandas.tseries.frequencies import to_offset
from xarray.core.dataarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_types import FrequencyLike, Indexer
from icclim.models.constants import (
    AMJJAS_MONTHS,
    DJF_MONTHS,
    EN_FREQ_MAPPING,
    FREQ_DELTA_MAPPING,
    JJA_MONTHS,
    MAM_MONTHS,
    MONTHS_MAP,
    ONDJFM_MONTHS,
    SON_MONTHS,
)
from icclim.models.registry import Registry
from icclim.utils import read_date

SEASON_ERR_MSG = (
    "A season created using `slice_mode` must be made of either"
    " consecutive integers for months such as [1,2,3] or two date strings"
    " such as ['19 july', '14 august']."
)

# RUN_INDEXER is a special value used for group by when there is no proper groupby to do
# but instead a filtering should be applied before the reducer.
RUN_INDEXER = "run_indexer"


def get_seasonal_time_updater(
    start_month: int, end_month: int, start_day: int = 1, end_day: int = None
) -> Callable[[DataArray], tuple[DataArray, DataArray]]:
    """Seasonal time updater and time bounds creator method generator.
    Returns a callable of DataArray which will rewrite the time dimension to
    the season composed of the given months. The data must have been computed on this
    season beforehand.
    It also create the corresponding time_bounds.

    Parameters
    ----------
    start_month: int
        The season starting month, it must be between 1 and 12.
    end_month: int
        The season ending month, it must be between 1 and 12.

    Returns
    -------
    function: Callable[[DataArray], DataArray]
        function resampling the input da to the wanted season.
    """

    def add_time_bounds(da: DataArray) -> tuple[DataArray, DataArray]:
        da_years = np.unique(da.time.dt.year)
        time_bounds = []
        new_time_axis = []
        first_time = da.time.values[0]
        for year in da_years:
            if start_month > end_month:
                year_of_season_end = year + 1
            else:
                year_of_season_end = year
            if isinstance(first_time, cftime.datetime):
                start = cftime.datetime(
                    year, start_month, start_day, calendar=first_time.calendar
                )
                end = _get_end_date(
                    use_cftime=True,
                    year=year_of_season_end,
                    month=end_month,
                    day=end_day,
                    calendar=first_time.calendar,
                )
            else:
                start = pd.to_datetime(f"{year}-{start_month}-{start_day}")
                end = _get_end_date(
                    use_cftime=False,
                    year=year_of_season_end,
                    month=end_month,
                    day=end_day,
                )
            new_time_axis.append(start + (end - start) / 2)
            time_bounds.append([start, end])
        da.coords["time"] = ("time", new_time_axis)
        time_bounds_da = DataArray(
            data=time_bounds,
            dims=["time", "bounds"],
            coords=[("time", da.time.values), ("bounds", [0, 1])],
        )
        return da, time_bounds_da

    return add_time_bounds


def get_time_bounds_updater(
    freq: str,
) -> Callable[[DataArray], tuple[DataArray, DataArray]]:
    def add_time_bounds(da: DataArray) -> tuple[DataArray, DataArray]:
        # da should already be resampled to freq
        if isinstance(da.indexes.get("time"), xr.CFTimeIndex):
            offset = xr.coding.cftime_offsets.to_offset(freq)
            starts = np.array(
                [
                    cftime.datetime(
                        date.year,
                        date.month,
                        date.day,
                        date.hour,
                        date.minute,
                        date.second,
                        calendar=date.calendar,
                    )
                    for date in da.indexes.get("time")
                ]
            )
            ends = starts + offset
            ends = ends - timedelta(days=1)
        else:
            offset = pd.tseries.frequencies.to_offset(freq)
            starts = pd.to_datetime(da.time.dt.floor("D"))
            ends = starts + offset
            ends = ends - pd.Timedelta(days=1)
        # make time axis values be in the middle of the bounds
        da["time"] = starts + (ends - starts) / 2
        time_bounds_da = DataArray(
            data=list(zip(starts, ends)),
            dims=["time", "bounds"],
            coords=[("time", da.time.values), ("bounds", [0, 1])],
        )
        return da, time_bounds_da

    return add_time_bounds


@dataclasses.dataclass
class Frequency:
    """Time sampling frequency."""

    pandas_freq: str
    accepted_values: list[str]
    adjective: str
    post_processing: Callable[[DataArray], tuple[DataArray, DataArray]] | None
    units: str
    indexer: Indexer | None
    long_name: str
    group_by_key: str | None
    delta: timedelta | np.timedelta64

    def build_frequency_kwargs(self) -> dict[str, Any]:
        """Build kwargs with possible keys in {"freq", "month", "date_bounds"}"""
        kwargs = dict(freq=self.pandas_freq)
        if self.indexer is not None:
            kwargs.update(self.indexer)
        return kwargs


class FrequencyRegistry(Registry[Frequency]):
    _item_class = Frequency

    HOUR = Frequency(
        pandas_freq="H",
        accepted_values=["hour", "h", "hourly"],
        adjective="hourly",
        indexer=None,
        post_processing=get_time_bounds_updater("H"),
        units="hours",
        long_name="hour",
        group_by_key="time.hour",
        delta=np.timedelta64(1, "h"),
    )
    """Resample to hourly values"""

    DAY = Frequency(
        pandas_freq="D",
        accepted_values=["daily", "day", "days", "d"],
        adjective="daily",
        indexer=None,
        post_processing=get_time_bounds_updater("D"),
        units="days",
        long_name="day",
        group_by_key="time.dayofyear",
        delta=np.timedelta64(1, "D"),
    )
    """Resample to daily values"""

    MONTH = Frequency(
        pandas_freq="MS",
        accepted_values=["month", "monthly", "MS"],
        adjective="monthly",
        indexer=None,
        post_processing=get_time_bounds_updater("MS"),
        units="months",
        long_name="month",
        group_by_key="time.month",
        delta=np.timedelta64(1, "M"),
    )
    """Resample to monthly values"""

    YEAR = Frequency(
        pandas_freq="YS",
        accepted_values=["year", "yearly", "annual", "YS"],
        adjective="annual",
        indexer=None,
        post_processing=get_time_bounds_updater("YS"),
        units="years",
        long_name="year",
        group_by_key="time.year",
        delta=np.timedelta64(1, "Y"),
    )
    """Resample to yearly values."""

    AMJJAS = Frequency(
        pandas_freq="AS-APR",
        accepted_values=["AMJJAS"],
        adjective="AMJJAS summery",
        indexer=dict(month=AMJJAS_MONTHS),
        post_processing=get_seasonal_time_updater(AMJJAS_MONTHS[0], AMJJAS_MONTHS[-1]),
        units="half_year_summers",
        long_name="AMJJAS season",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(6, "M"),
    )
    """Resample to summer half-year, from April to September included."""

    ONDJFM = Frequency(
        pandas_freq="AS-OCT",
        accepted_values=["ONDJFM"],
        adjective="ONDJFM wintry",
        indexer=dict(month=ONDJFM_MONTHS),
        post_processing=get_seasonal_time_updater(ONDJFM_MONTHS[0], ONDJFM_MONTHS[-1]),
        units="half_year_winters",
        long_name="ONDJFM season",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(6, "M"),
    )
    """Resample to winter half-year, from October to March included."""

    DJF = Frequency(
        pandas_freq="AS-DEC",
        accepted_values=["DJF"],
        adjective="DJF wintry",
        indexer=dict(month=DJF_MONTHS),
        post_processing=get_seasonal_time_updater(DJF_MONTHS[0], DJF_MONTHS[-1]),
        units="winters",
        long_name="DJF winter",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(3, "M"),
    )
    """Resample to winter season, from December to February included."""

    MAM = Frequency(
        pandas_freq="AS-MAR",
        accepted_values=["MAM"],
        adjective="MAM springlong",
        indexer=dict(month=MAM_MONTHS),
        post_processing=get_seasonal_time_updater(MAM_MONTHS[0], MAM_MONTHS[-1]),
        units="springs",
        long_name="MAM season",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(3, "M"),
    )
    """Resample to spring season, from March to May included."""

    JJA = Frequency(
        pandas_freq="AS-JUN",
        accepted_values=["JJA"],
        adjective="JJA summery",
        indexer=dict(month=JJA_MONTHS),
        post_processing=get_seasonal_time_updater(JJA_MONTHS[0], JJA_MONTHS[-1]),
        units="summers",
        long_name="JJA season",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(3, "M"),
    )
    """Resample to summer season, from June to Agust included."""

    SON = Frequency(
        pandas_freq="AS-SEP",
        accepted_values=["SON"],
        adjective="SON autumnal",
        indexer=dict(month=SON_MONTHS),
        post_processing=get_seasonal_time_updater(SON_MONTHS[0], SON_MONTHS[-1]),
        units="autumns",
        long_name="SON season",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(3, "M"),
    )
    """Resample to fall season, from September to November included."""

    @classmethod
    def lookup(cls, item: FrequencyLike, no_error: bool = False) -> Frequency | None:
        if isinstance(item, Frequency):
            return item
        if isinstance(item, str):
            return _get_frequency_from_string(item)
        if isinstance(item, (list, tuple)):
            return _get_frequency_from_iterable(item)
        if no_error:
            return None
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {item}."
            f" Use a Frequency from {FrequencyRegistry.every_aliases()}"
        )

    @staticmethod
    def get_item_aliases(item: Frequency) -> list[str]:
        return item.accepted_values


def _get_end_date(
    use_cftime: bool, year: int, month: int, day: int = None, calendar=None
):
    delta = timedelta(days=0)
    if day is None:
        if month == 12:
            day = 31
        else:
            # get the next month and subtract a day (handle any month and leap years)
            month = month + 1
            day = 1
            delta = timedelta(days=1)
    if use_cftime:
        end = cftime.datetime(year, month, day, calendar=calendar)
    else:
        end = pd.to_datetime(f"{year}-{month}-{day}")
    return end - delta


def _get_frequency_from_string(query: str) -> Frequency:
    for key, freq in FrequencyRegistry.catalog().items():
        if key == query.upper() or query.upper() in map(
            str.upper, freq.accepted_values
        ):
            return freq
    # else assumes it's a pandas frequency (such as "W" or "3MS")
    try:
        to_offset(query)  # no-op, used to check if it's a valid pandas freq
    except ValueError as e:
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {query}. Use either a"
            " valid icclim frequency or a valid pandas"
            " frequency",
            e,
        )
    return Frequency(
        post_processing=get_time_bounds_updater(query),
        pandas_freq=query,
        adjective=f"time series sampled on {query}",
        accepted_values=[],
        indexer=None,
        units=query,
        long_name=_get_long_name(query),
        group_by_key=None,
        delta=_get_delta(query),
    )


def _is_season_valid(months: list[int]) -> bool:
    is_valid = True
    for i in range(0, len(months) - 1):
        is_valid = is_valid and 0 < months[i] < 13
        if months[i] > months[i + 1]:
            is_valid = is_valid and months[i + 1] == 1 and months[i] == 12
        else:
            is_valid = is_valid and (months[i + 1] - months[i] == 1)
    return is_valid


def _get_frequency_from_iterable(
    slice_mode_list: list | tuple[str, Sequence]
) -> Frequency:
    if len(slice_mode_list) < 2:
        raise InvalidIcclimArgumentError(
            "Invalid slice_mode format."
            " When slice_mode is a list, its first element must be a keyword and"
            " its second a list (e.g `slice_mode=['season', [1,2,3]]` )."
        )
    freq_keyword = slice_mode_list[0]
    if freq_keyword in ["month", "months"]:
        return _build_frequency_filtered_by_month(slice_mode_list[1])
    elif freq_keyword in ["season", "seasons"]:
        season = slice_mode_list[1]
        return _build_seasonal_freq(season)
    else:
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {slice_mode_list}."
            " The sampling frequency must be one of {'season', 'month'}"
        )


def _build_frequency_filtered_by_month(months: Sequence[int]) -> Frequency:
    return Frequency(
        indexer=dict(month=months),
        post_processing=get_time_bounds_updater("MS"),
        pandas_freq="MS",
        adjective="monthly",
        accepted_values=[],
        units="months",
        long_name=f"monthly time series (months: {months})",
        group_by_key="time.month",
        delta=np.timedelta64(1, "M"),
    )


def _build_seasonal_freq(season: Sequence):
    if isinstance(season[0], str):
        return _build_seasonal_frequency_between_dates(season)
    elif isinstance(season, tuple) or isinstance(season[0], int):
        return _build_seasonal_frequency_for_months(season)
    else:
        raise NotImplementedError()


def _build_seasonal_frequency_between_dates(season: Sequence[str]) -> Frequency:
    if len(season) != 2:
        raise InvalidIcclimArgumentError(SEASON_ERR_MSG)
    begin_date = read_date(season[0])
    end_date = read_date(season[1])
    begin_formatted = begin_date.strftime("%m-%d")
    end_formatted = end_date.strftime("%m-%d")
    indexer = dict(date_bounds=(begin_formatted, end_formatted))
    return Frequency(
        indexer=indexer,
        post_processing=get_seasonal_time_updater(
            begin_date.month, end_date.month, begin_date.day, end_date.day
        ),
        pandas_freq=f"AS-{MONTHS_MAP[begin_date.month]}",
        adjective="seasonally",
        accepted_values=[],
        units=f"{MONTHS_MAP[begin_date.month]}_{MONTHS_MAP[end_date.month]}_seasons",
        long_name=f"seasonal time series"
        f" (season: from {begin_formatted} to {end_formatted})",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(end_date - begin_date),
    )


def _build_seasonal_frequency_for_months(season: tuple | list):
    if isinstance(season, tuple):
        # concat in case of ([12], [1, 2])
        season = season[0] + season[1]
    if not _is_season_valid(season):
        raise InvalidIcclimArgumentError(SEASON_ERR_MSG)
    indexer = dict(month=season)
    return Frequency(
        indexer=indexer,
        post_processing=get_seasonal_time_updater(season[0], season[-1]),
        pandas_freq=f"AS-{MONTHS_MAP[season[0]]}",
        adjective="seasonally",
        accepted_values=[],
        units=f"{MONTHS_MAP[season[0]]}_{MONTHS_MAP[season[-1]]}_seasons",
        long_name=f"seasonal time series (season: {season})",
        group_by_key=RUN_INDEXER,
        delta=np.timedelta64(len(season), "M"),
    )


def _get_long_name(pandas_freq: str) -> str:
    no_digit_freq = re.findall(r"\D+", pandas_freq)[0]
    multiplier = re.findall(r"\d+", pandas_freq)
    freqs = no_digit_freq.split("-")[::-1]  # reverse
    freqs = [EN_FREQ_MAPPING[f] for f in freqs]
    freqs = " ".join(freqs)
    if multiplier:
        return f"{multiplier[0]} {freqs}"
    else:
        return freqs


def _get_delta(pandas_freq: str) -> np.timedelta64:
    """
    Build timedelta from a "pandas frequency" string.
    A "pandas frequency" string may look like ["2AS-DEC", "3W-TUE", "M", ... ]
    The anchor, such as "DEC" in "AS-DEC", does not modify the delta.

    Parameters
    ----------
    pandas_freq : str
    The frequency query.

    Returns
    -------
    The timedelta corresponding to this frequency.
    For example, "2AS-DEC" would return a 2 years delta.
    """
    # [0] to ignore the anchor
    non_digit = re.findall(r"\D+", pandas_freq)[0].split("-")[0]
    base, freq = FREQ_DELTA_MAPPING[non_digit]
    # we assume the starting digits are the multiplier.
    multiplier = re.findall(r"\d+", pandas_freq)
    if multiplier:
        multiplier = int(multiplier[0])
        return np.timedelta64(base * multiplier, freq)
    else:
        return np.timedelta64(base, freq)



from __future__ import annotations

import os
import shutil

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from xclim.core.utils import PercentileDataArray

from icclim.ecad.ecad_indices import EcadIndexRegistry
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import UNITS_KEY
from icclim.pre_processing.input_parsing import (
    guess_var_names,
    read_dataset,
    update_to_standard_coords,
)


def test_update_to_standard_coords():
    # GIVEN
    ds = xr.Dataset(
        {
            "pouet": xr.DataArray(
                data=np.full(10, 42).reshape((10, 1, 1)),
                coords=dict(
                    latitude=[42],
                    longitude=[42],
                    t=pd.date_range("2042-01-01", periods=10, freq="D"),
                ),
                dims=["t", "latitude", "longitude"],
                name="pr",
                attrs={UNITS_KEY: "kg m-2 d-1"},
            )
        }
    )
    # WHEN
    res = update_to_standard_coords(ds)
    # THEN
    assert "time" in res.coords


class Test_ReadDataset:
    OUTPUT_NC_FILE = "tmp.nc"
    OUTPUT_NC_FILE_2 = "tmp-2.nc"
    OUTPUT_ZARR_STORE = "tmp.zarr"
    OUTPUT_UNKNOWN_FORMAT = "tmp.cacahuete"
    pr_da = None
    tas_da = None

    @pytest.fixture(autouse=True)
    def cleanup(self):
        # -- setup
        self.pr_da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                time=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["time", "latitude", "longitude"],
            name="pr",
            attrs={UNITS_KEY: "kg m-2 d-1"},
        )
        self.tas_da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                t=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["t", "latitude", "longitude"],
            name="tas",
            attrs={UNITS_KEY: "degC"},
        )
        yield
        # -- teardown
        shutil.rmtree(self.OUTPUT_ZARR_STORE, ignore_errors=True)
        for f in [
            self.OUTPUT_NC_FILE,
            self.OUTPUT_NC_FILE_2,
            self.OUTPUT_UNKNOWN_FORMAT,
        ]:
            try:
                os.remove(f)
            except FileNotFoundError:
                pass

    def test_read_dataset_xr_DataArray__simple(self):
        # WHEN
        res = read_dataset(self.pr_da)
        # THEN
        assert "pr" in res.data_vars

    def test_read_dataset_xr_da_user_index_success(self):
        # WHEN
        ds_res = read_dataset(self.pr_da, None)
        # THEN
        xr.testing.assert_equal(ds_res.pr, self.pr_da)

    def test_read_dataset_xr_ds__simple(self):
        # GIVEN
        ds = xr.Dataset({"pouet": self.pr_da})
        # WHEN
        ds_res = read_dataset(ds)
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)

    def test_read_dataset_netcdf_success(self):
        # GIVEN
        ds = xr.Dataset({"pouet": self.pr_da})
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        # WHEN
        ds_res = read_dataset(self.OUTPUT_NC_FILE)
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)

    def test_read_dataset_multi_netcdf_success(self):
        # GIVEN
        ds = xr.Dataset({"pouet": self.pr_da})
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        ds.rename({"pouet": "patapouet"}).to_netcdf(self.OUTPUT_NC_FILE_2)
        # WHEN
        ds_res = read_dataset([self.OUTPUT_NC_FILE, self.OUTPUT_NC_FILE_2])
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)
        xr.testing.assert_equal(ds_res.patapouet, ds.pouet)

    def test_read_dataset_zarr_store_success(self):
        # GIVEN
        ds = xr.Dataset({"pouet": self.pr_da})
        ds.to_zarr(self.OUTPUT_ZARR_STORE)
        # WHEN
        ds_res = read_dataset(self.OUTPUT_ZARR_STORE)
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)

    def test_read_dataset_not_implemented_error(self):
        # THEN
        with pytest.raises(NotImplementedError):
            # WHEN
            read_dataset(42)  # noqa

    def test_read_dataset(self):
        # GIVEN
        ds = xr.Dataset({"tas": self.tas_da})
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        # WHEN
        res_ds = read_dataset(self.OUTPUT_NC_FILE)
        # THEN
        # asserts variable names are the ones in the actual DataArray/Datasets
        assert "tas" in res_ds.data_vars

    def test_read_dataset__with_percentiles(self):
        # GIVEN
        ds = xr.Dataset({"tas": self.tas_da})
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        per = self.tas_da
        per.coords["percentiles"] = 42
        per = per.rename("tontontonthetatilotetatoux").expand_dims("percentiles")
        per = PercentileDataArray.from_da(
            per, climatology_bounds=["1994-12-02", "1999-01-01"]
        )
        ds["tontontonthetatilotetatoux"] = per
        # WHEN
        res_ds = read_dataset(ds)
        # THEN
        assert "tas" in res_ds.data_vars
        assert "tontontonthetatilotetatoux" in res_ds.data_vars

    def test_guess_variables__cant_guess_var_name(self):
        # GIVEN
        ds = xr.Dataset({"canard": self.tas_da, "bergeronnette": self.tas_da})
        # THEN
        with pytest.raises(InvalidIcclimArgumentError):
            # WHEN
            guess_var_names(ds, standard_index=EcadIndexRegistry.SU, var_names=None)

    def test_guess_variables__simple(self):
        # GIVEN
        ds = xr.Dataset({"tas": self.tas_da})
        # WHEN
        res = guess_var_names(ds, standard_index=EcadIndexRegistry.TG, var_names=None)
        # THEN
        assert res == ["tas"]

    def test_guess_variables__from_string(self):
        # GIVEN
        ds = xr.Dataset({"tas": self.tas_da})
        # WHEN
        res = guess_var_names(ds, standard_index=None, var_names="cocoLasticot")
        # THEN
        assert res == ["cocoLasticot"]

    def test_guess_variables__from_list(self):
        # GIVEN
        ds = xr.Dataset({"tas": self.tas_da})
        # WHEN
        res = guess_var_names(ds, standard_index=None, var_names=["pinçon"])
        # THEN
        assert res == ["pinçon"]

    def test_guess_variables__from_alias(self):
        # GIVEN
        ds = xr.Dataset({"tasmaxAdjust": self.tas_da, "turlututut": self.tas_da})
        # WHEN
        res = guess_var_names(ds, standard_index=EcadIndexRegistry.SU, var_names=None)
        # THEN
        assert res == ["tasmaxAdjust"]



from __future__ import annotations


class InvalidIcclimArgumentError(ValueError):
    """Exceptions raised erroneous input arguments.

    Parameters
    ----------
        msg: str
            Error description.
        source_err: Exception | None
            The source of the error if any.
    """

    def __init__(self, msg: str, source_err: Exception = None):
        self.msg = msg
        self.source = source_err

    def __str__(self):
        return repr(self.msg)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Sequence

import icclim.models.index_group as index_group
from icclim.generic_indices.standard_variable import StandardVariable


@dataclass
class StandardIndex:
    """Standard Index data class.
    It is used to describe how a GenericIndicator should be setup to compute a climate
    index that has been defined in the literature (such as ECA&D's ATBD document).


    Attributes
    ----------

    short_name: str
        The index name used in the output.
    compute: Callable
        The function to compute the index. It usually wraps a xclim functions.
    group: IndexGroup
        The index group category.
    variables: List[List[str]]
        The Cf variables needed to compute the index.
        The variable are individually described by a list of aliases.
    qualifiers: List[str] | None
        ``optional`` List of configuration to compute the index.
        Used internally to generate modules for C3S.
    source: str | None
        Where the index definition comes from.
    definition: str | None
        A formal definition of the index. It should describe what kind of output
        the user is expected to obtain.
    """

    short_name: str
    group: index_group.IndexGroup
    input_variables: list[StandardVariable] | None  # None when index is generic
    indicator: Any  # Any -> Indicator (circular dep)
    # todo: merge qualifiers with group into a Set of qualifiers ?
    qualifiers: list[str] | None = None
    source: str | None = None
    reference: str | None = None
    definition: str | None = None
    threshold: str | None | Any | Sequence[str | Any] = None  # Any -> Threshold
    output_unit: str | None = None
    # additional, index specific args
    rolling_window_width: int | None = None
    doy_window_width: int | None = None
    min_spell_length: int | None = None

    def __str__(self):
        return f"{self.group} | {self.short_name} | {self.definition}"

    def __call__(self, *args, **kwargs):
        self.indicator(*args, **kwargs)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, StandardIndex):
            return False
        return (
            self.indicator == other.indicator
            and self.threshold == other.threshold
            and self.output_unit == other.output_unit
            and self.rolling_window_width == other.rolling_window_width
            and self.doy_window_width == other.doy_window_width
            and self.min_spell_length == other.min_spell_length
        )



from __future__ import annotations

import dataclasses
from functools import reduce
from typing import Callable

import numpy as np
from xarray import DataArray

from icclim.models.registry import Registry


@dataclasses.dataclass
class LogicalLink:
    name: str
    standard_name: str
    long_name: str
    short_name: str
    compute: Callable[[list[DataArray]], DataArray]

    def __call__(self, *args, **kwargs) -> DataArray:
        return self.compute(*args, **kwargs)


class LogicalLinkRegistry(Registry[LogicalLink]):
    _item_class = LogicalLink

    LOGICAL_OR = LogicalLink(
        name="or",
        standard_name="OR",
        long_name="OR",
        short_name="OR",
        compute=lambda data_list: reduce(np.logical_or, data_list),  # type:ignore
    )
    LOGICAL_AND = LogicalLink(
        name="and",
        standard_name="AND",
        long_name="AND",
        short_name="AND",
        compute=lambda data_list: reduce(np.logical_and, data_list),  # type:ignore
    )



from __future__ import annotations

import re
import warnings
from datetime import datetime

import dateparser
import pint
import xarray as xr
import xclim

from icclim.icclim_exceptions import InvalidIcclimArgumentError

PR_AMOUNT_STANDARD_NAME = "thickness_of_rainfall_amount"


def read_date(in_date: str | datetime) -> datetime:
    if isinstance(in_date, datetime):
        return in_date
    date = dateparser.parse(in_date)
    if date is None:
        raise InvalidIcclimArgumentError(
            f"The date {in_date} does not have a valid format."
            " You can use various formats such as '2 december', '02-12',"
            " '1994-12-02'..."
        )
    return date


def get_date_to_iso_format(in_date: str | datetime) -> str:
    if isinstance(in_date, str):
        if re.match(r"^\d{4}$", in_date):
            warnings.warn(f"{in_date} is transformed into {in_date}-01-01")
            in_date += "-01-01"
        if re.match(r"^\d{4}-\d{2}$", in_date):
            warnings.warn(f"{in_date} is transformed into {in_date}-01")
            in_date += "-01"
        in_date = read_date(in_date)
    return in_date.strftime("%Y-%m-%d")


def is_number_sequence(values) -> bool:
    return isinstance(values, (tuple, list)) and all(
        map(lambda x: isinstance(x, (float, int)), values)
    )


def _is_rate(u: pint.Unit) -> bool:
    return u.dimensionality.get("[time]") == -1


def _is_amount(u: pint.Unit) -> bool:
    return not _is_rate(u)


def is_precipitation_amount(source: xr.DataArray) -> bool:
    standard_name = source.attrs.get("standard_name", None)
    source_unit = xclim.core.units.units2pint(source)
    return standard_name == PR_AMOUNT_STANDARD_NAME and _is_amount(source_unit)



from __future__ import annotations

import dataclasses
from typing import Literal

from icclim.models.registry import Registry


@dataclasses.dataclass
class NetcdfVersion:
    name: Literal["NETCDF4", "NETCDF4_CLASSIC", "NETCDF3_64BIT", "NETCDF3_CLASSIC"]


class NetcdfVersionRegistry(Registry[NetcdfVersion]):
    _item_class = NetcdfVersion

    NETCDF4 = NetcdfVersion("NETCDF4")
    NETCDF4_CLASSIC = NetcdfVersion("NETCDF4_CLASSIC")
    NETCDF3_CLASSIC = NetcdfVersion("NETCDF3_CLASSIC")
    NETCDF3_64BIT = NetcdfVersion("NETCDF3_64BIT")



from __future__ import annotations

import xarray
import xclim

from icclim.generic_indices.generic_indicators import (
    Indicator,
    get_couple_of_var,
    get_single_var,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.frequency import FrequencyRegistry
from icclim.models.index_config import IndexConfig


class XCLIM_BINDING:
    class GrowingSeasonLength(Indicator):
        """
        Fake icclim indicator that redirect to xclim `growing_season_length` indicator.
        """

        name = "growing_season_length"
        standard_name = xclim.atmos.growing_season_length.standard_name
        long_name = "ECAD Growing Season Length (Tmean > 5 degree_Celsius)"
        cell_methods = ""

        def __call__(self, config: IndexConfig) -> xarray.DataArray:
            study, threshold = get_single_var(config.climate_variables)
            return xclim.atmos.growing_season_length(
                tas=study,
                thresh="5 degree_Celsius",
                window=6,
                mid_date="07-01",
                freq=config.frequency.pandas_freq,
            )

        def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()

        def postprocess(self, *args, **kwargs) -> xarray.DataArray:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()

    class StandardizedPrecipitationIndex3(Indicator):
        """
        Fake icclim indicator that redirect to xclim `standardized_precipitation_index`
        indicator, with 3 MS configured.
        """

        name = "standardized_precipitation_index_3"
        standard_name = xclim.atmos.standardized_precipitation_index.standard_name
        long_name = "3-Month Standardized Precipitation Index (SPI3)"
        cell_methods = ""

        def __call__(self, config: IndexConfig) -> xarray.DataArray:
            if config.frequency is not FrequencyRegistry.YEAR:  # year is default freq
                raise InvalidIcclimArgumentError(
                    "`slice_mode` cannot be configured when computing SPI3"
                )
            study, ref = get_couple_of_var(config.climate_variables, "SPI")
            return xclim.atmos.standardized_precipitation_index(
                pr=study, pr_cal=ref, freq="MS", window=3, dist="gamma", method="APP"
            )

        def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()

        def postprocess(self, *args, **kwargs) -> xarray.DataArray:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()

    class StandardizedPrecipitationIndex6(Indicator):
        """
        Fake icclim indicator that redirect to xclim `standardized_precipitation_index`
        indicator, with 6 MS configured.
        """

        name = "standardized_precipitation_index_6"
        standard_name = xclim.atmos.standardized_precipitation_index.standard_name
        long_name = "6-Month Standardized Precipitation Index (SPI6)"
        cell_methods = ""

        def __call__(self, config: IndexConfig) -> xarray.DataArray:
            if config.frequency is not FrequencyRegistry.YEAR:  # year is default freq
                raise InvalidIcclimArgumentError(
                    "`slice_mode` cannot be configured when computing SPI6"
                )
            study, ref = get_couple_of_var(config.climate_variables, "SPI")
            return xclim.atmos.standardized_precipitation_index(
                pr=study, pr_cal=ref, freq="MS", window=6, dist="gamma", method="APP"
            )

        def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()

        def postprocess(self, *args, **kwargs) -> xarray.DataArray:
            """Not implemented as xclim indicator already handle pre/post processing"""
            raise NotImplementedError()



from __future__ import annotations

import dataclasses

from icclim.models.registry import Registry


@dataclasses.dataclass
class QuantileInterpolation:
    name: str
    alpha: float
    beta: float


class QuantileInterpolationRegistry(Registry[QuantileInterpolation]):
    _item_class = QuantileInterpolation

    LINEAR = QuantileInterpolation("linear", 1, 1)
    MEDIAN_UNBIASED = QuantileInterpolation("median_unbiased", 1.0 / 3, 1.0 / 3)



from __future__ import annotations

import pytest

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.index_group import IndexGroupRegistry


@pytest.mark.parametrize(
    "gr",
    [
        ("temperature", IndexGroupRegistry.TEMPERATURE),
        ("heat", IndexGroupRegistry.HEAT),
        ("cold", IndexGroupRegistry.COLD),
        ("drought", IndexGroupRegistry.DROUGHT),
        ("rain", IndexGroupRegistry.RAIN),
        ("snow", IndexGroupRegistry.SNOW),
    ],
)
def test_lookup_success(gr):
    assert IndexGroupRegistry.lookup(gr[0]) == gr[1]


def test_lookup_error():
    with pytest.raises(InvalidIcclimArgumentError):
        IndexGroupRegistry.lookup("coin coin le canard")


def test_union():
    x = IndexGroupRegistry.RAIN | IndexGroupRegistry.SNOW
    assert x.name == "rain_snow"
    assert x.values[0] == IndexGroupRegistry.RAIN
    assert x.values[1] == IndexGroupRegistry.SNOW



# -*- Coding: latin-1 -*-
#  Copyright CERFACS (http://cerfacs.fr/)
#  Apache License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
"""
Main entry point of icclim.
This module expose icclim principal function, notably `index` which is use by the
generated API.
A convenience function `indices` is also exposed to compute multiple indices at once.
"""
from __future__ import annotations

import time
from datetime import datetime
from functools import partial, reduce
from typing import Callable, Literal, Sequence
from warnings import warn

import xarray as xr
import xclim
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

from icclim.ecad.ecad_indices import EcadIndexRegistry
from icclim.generic_indices.generic_indicators import (
    GenericIndicator,
    GenericIndicatorRegistry,
    Indicator,
)
from icclim.generic_indices.threshold import Threshold, build_threshold
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_logger import IcclimLogger, Verbosity, VerbosityRegistry
from icclim.icclim_types import InFileLike, SamplingMethodLike
from icclim.models.climate_variable import (
    ClimateVariable,
    build_climate_vars,
    read_in_files,
)
from icclim.models.constants import (
    ICCLIM_VERSION,
    PERCENTILE_THRESHOLD_STAMP,
    RESAMPLE_METHOD,
    UNITS_KEY,
    USER_INDEX_PRECIPITATION_STAMP,
    USER_INDEX_TEMPERATURE_STAMP,
)
from icclim.models.frequency import Frequency, FrequencyLike, FrequencyRegistry
from icclim.models.index_config import IndexConfig
from icclim.models.index_group import IndexGroup, IndexGroupRegistry
from icclim.models.logical_link import LogicalLink, LogicalLinkRegistry
from icclim.models.netcdf_version import NetcdfVersion, NetcdfVersionRegistry
from icclim.models.operator import Operator, OperatorRegistry
from icclim.models.quantile_interpolation import (
    QuantileInterpolation,
    QuantileInterpolationRegistry,
)
from icclim.models.standard_index import StandardIndex
from icclim.models.user_index_dict import UserIndexDict
from icclim.pre_processing.in_file_dictionary import InFileDictionary
from icclim.user_indices.calc_operation import CalcOperationRegistry
from icclim.utils import read_date

log: IcclimLogger = IcclimLogger.get_instance(VerbosityRegistry.LOW)

HISTORY_CF_KEY = "history"
SOURCE_CF_KEY = "source"
ICCLIM_REFERENCE = "icclim"


def indices(
    index_group: Literal["all"] | str | IndexGroup | Sequence[str],
    ignore_error: bool = False,
    **kwargs,
) -> Dataset:
    """
    Compute multiple indices at the same time.
    The input dataset(s) must include all the necessary variables.
    It can only be used with keyword arguments (kwargs)

    .. notes
        If ``output_file`` is part of kwargs, the result is written in a single netCDF
        file, which will contain all the index results of this group.

    Parameters
    ----------
    index_group : "all" | str | IndexGroup | list[str]
        Either the name of an IndexGroup or an instance of IndexGroup or a list
        of index short names or the name(s) of standard variable(s) (such as 'tasmax').
        The value "all" can also be used to compute every indices.
        Note that the input given by ``in_files`` must include all the necessary
        variables to compute the indices of this group.
    ignore_error: bool
        When True, ignore indices that fails to compute. This is option is particularly
        useful when used with `index_group='all'` to compute everything that can be
        computed given the input.
    kwargs : Dict
        ``icclim.index`` keyword arguments.

    Returns
    -------
    xr.Dataset
        A Dataset with one data variable per index.

    """
    indices = _get_indices_of_group(index_group)
    out = None
    if "out_file" in kwargs.keys():
        out = kwargs["out_file"]
        del kwargs["out_file"]
    acc = []
    for i in indices:
        log.info(f"Computing index '{i.short_name}'")
        kwargs["index_name"] = i.short_name
        if ignore_error:
            try:
                res = index(**kwargs)
                if "percentiles" in res.coords:
                    res = res.rename({"percentiles": i.short_name + "_percentiles"})
                if "thresholds" in res.coords:
                    res = res.rename({"thresholds": i.short_name + "_thresholds"})
                acc.append(res)
            except Exception:
                warn(f"Could not compute {i.short_name}.")
        else:
            res = index(**kwargs)
            if "percentiles" in res.coords:
                res = res.rename({"percentiles": i.short_name + "_percentiles"})
            if "thresholds" in res.coords:
                res = res.rename({"thresholds": i.short_name + "_thresholds"})
            acc.append(res)
    ds: Dataset = xr.merge(acc)
    if out is not None:
        _write_output_file(
            result_ds=ds,
            input_time_encoding=ds.time.encoding,
            netcdf_version=kwargs.get("netcdf_version", NetcdfVersionRegistry.NETCDF4),
            file_path=out,
        )
    return ds


def _get_indices_of_group(
    query: list | tuple | str | IndexGroup | StandardIndex,
) -> list[StandardIndex]:
    if query == IndexGroupRegistry.WILD_CARD_GROUP or (
        isinstance(query, str)
        and query.lower() == IndexGroupRegistry.WILD_CARD_GROUP.name
    ):
        # case of group="all"
        return EcadIndexRegistry.values()
    if not isinstance(query, (list, tuple)):
        query = [query]
    # -- Look for standard indices (e.g. index_group='tx90p')
    indices = [EcadIndexRegistry.lookup(i, no_error=True) for i in query]
    indices = list(filter(lambda x: x is not None, indices))
    if len(indices) == len(query):
        return indices
    # -- Look for variables in standard indices (e.g. index_group='tasmax')
    indices = []
    for ecad_index in EcadIndexRegistry.values():
        has_var = True
        for var in ecad_index.input_variables:
            is_query_in_aliases = map(
                lambda standard_var: standard_var in var.aliases, query
            )
            has_var &= any(is_query_in_aliases)
        if has_var:
            indices.append(ecad_index)
    if len(indices) >= len(query):
        return indices
    # -- Look for index group (e.g. index_group='HEAT')
    groups = [IndexGroupRegistry.lookup(i, no_error=True) for i in query]
    groups = list(filter(lambda x: x is not None, groups))
    indices = map(lambda x: x.get_indices(), groups)
    indices = reduce(lambda x, y: x + y, indices, [])  # flatten list[list]
    if len(indices) >= len(query):
        return indices
    raise InvalidIcclimArgumentError(f"The index group {query} was not recognized.")


def indice(*args, **kwargs) -> Dataset:
    """
    Deprecated proxy for `icclim.index` function.
    To be deleted in a futur version.
    """
    log.deprecation_warning(old="icclim.indice", new="icclim.index")
    return index(*args, **kwargs)


def index(
    in_files: InFileLike,
    index_name: str | None = None,  # optional when computing user_indices
    var_name: str | Sequence[str] | None = None,
    slice_mode: FrequencyLike | Frequency = "year",
    time_range: Sequence[datetime | str] | None = None,
    out_file: str | None = None,
    threshold: str | Threshold | Sequence[str | Threshold] = None,
    callback: Callable[[int], None] = log.callback,
    callback_percentage_start_value: int = 0,
    callback_percentage_total: int = 100,
    base_period_time_range: Sequence[datetime] | Sequence[str] | None = None,
    doy_window_width: int = 5,
    only_leap_years: bool = False,
    ignore_Feb29th: bool = False,
    interpolation: str | QuantileInterpolation = "median_unbiased",
    out_unit: str | None = None,
    netcdf_version: str | NetcdfVersion = "NETCDF4",
    user_index: UserIndexDict | None = None,
    save_thresholds: bool = False,
    logs_verbosity: Verbosity | str = "LOW",
    date_event: bool = False,
    min_spell_length: int | None = 6,
    rolling_window_width: int | None = 5,
    sampling_method: SamplingMethodLike = RESAMPLE_METHOD,
    *,
    # deprecated params are kwargs only
    window_width: int | None = None,
    save_percentile: bool | None = None,
    indice_name: str = None,
    user_indice: UserIndexDict = None,
    transfer_limit_Mbytes: float = None,
) -> Dataset:
    """
    Main entry point for icclim to compute climate indices.

    Parameters
    ----------

    in_files: str | list[str] | Dataset | DataArray | InputDictionary
        Absolute path(s) to NetCDF dataset(s), including OPeNDAP URLs,
        or path to zarr store, or xarray.Dataset or xarray.DataArray.
    index_name: str
        Climate index name.
        For ECA&D index, case insensitive name used to lookup the index.
        For user index, it's the name of the output variable.
    var_name: str | list[str] | None
        ``optional`` Target variable name to process corresponding to ``in_files``.
        If None (default) on ECA&D index, the variable is guessed based on the climate
        index wanted.
        Mandatory for a user index.
    slice_mode: SliceMode
        Type of temporal aggregation:
        The possibles values are ``{"year", "month", "DJF", "MAM", "JJA", "SON",
        "ONDJFM" or "AMJJAS", ("season", [1,2,3]), ("month", [1,2,3,])}``
        (where season and month lists can be customized) or any valid pandas frequency.
        A season can also be defined between two exact dates:
        ``("season", ("19 july", "14 august"))``.
        Default is "year".
        See :ref:`slice_mode` for details.
    time_range: list[datetime ] | list[str]  | tuple[str, str] | None
        ``optional`` Temporal range: upper and lower bounds for temporal subsetting.
        If ``None``, whole period of input files will be processed.
        The dates can either be given as instance of datetime.datetime or as string
        values. For strings, many format are accepted.
        Default is ``None``.
    out_file: str | None
        Output NetCDF file name (default: "icclim_out.nc" in the current directory).
        Default is "icclim_out.nc".
        If the input ``in_files`` is a ``Dataset``, ``out_file`` field is ignored.
        Use the function returned value instead to retrieve the computed value.
        If ``out_file`` already exists, icclim will overwrite it!
    threshold: float | list[float] | None
        ``optional`` User defined threshold for certain indices.
        Default depend on the index, see their individual definition.
        When a list of threshold is provided, the index will be computed for each
        thresholds.
    transfer_limit_Mbytes: float
        Deprecated, does not have any effect.
    callback: Callable[[int], None]
        ``optional`` Progress bar printing. If ``None``, progress bar will not be
        printed.
    callback_percentage_start_value: int
        ``optional`` Initial value of percentage of the progress bar (default: 0).
    callback_percentage_total: int
        ``optional`` Total percentage value (default: 100).
    base_period_time_range: list[datetime ] | list[str]  | tuple[str, str] | None
        ``optional`` Temporal range of the reference period.
        The dates can either be given as instance of datetime.datetime or as string
        values.
        It is used either:
        #. to compute percentiles if threshold is filled.
        When missing, the studied period is used to compute percentiles.
        The study period is either the dataset filtered by `time_range` or the whole
        dataset if `time_range` is missing.
        For day of year percentiles (doy_per), on extreme percentiles the
        overlapping period between `base_period_time_range` and the study period is
        bootstrapped.
        #. to compute a reference period for indices such as difference_of_mean
        (a.k.a anomaly) if a single variable is given in input.
    doy_window_width: int
        ``optional`` Window width used to aggreagte day of year values when computing
        day of year percentiles (doy_per)
        Default: 5 (5 days).
    min_spell_length: int
        ``optional`` Minimum spell duration to be taken into account when computing the
        sum_of_spell_lengths.
    rolling_window_width: int
        ``optional`` Window width of the rolling window for indicators such as
        `{max_of_rolling_sum, max_of_rolling_average, min_of_rolling_sum, min_of_rolling_average}`  # noqa
    only_leap_years: bool
        ``optional`` Option for February 29th (default: False).
    ignore_Feb29th: bool
        ``optional`` Ignoring or not February 29th (default: False).
    interpolation: str | QuantileInterpolation | None
        ``optional`` Interpolation method to compute percentile values:
        ``{"linear", "median_unbiased"}``
        Default is "median_unbiased", a.k.a type 8 or method 8.
        Ignored for non percentile based indices.
    out_unit: str | None
        ``optional`` Output unit for certain indices: "days" or "%" (default: "days").
    netcdf_version: str | NetcdfVersion
        ``optional`` NetCDF version to create (default: "NETCDF3_CLASSIC").
    user_index: UserIndexDict
        ``optional`` A dictionary with parameters for user defined index.
        See :ref:`Custom indices`.
        Ignored for ECA&D indices.
    save_thresholds: bool
        ``optional`` True if the thresholds should be saved within the resulting netcdf
         file (default: False).
    date_event: bool
        When True the date of the event (such as when a maximum is reached) will be
        stored in coordinates variables.
        **warning** This option may significantly slow down computation.
    logs_verbosity: str | Verbosity
        ``optional`` Configure how verbose icclim is.
        Possible values: ``{"LOW", "HIGH", "SILENT"}`` (default: "LOW")
    sampling_method: str
        Choose whether the output sampling configured in `slice_mode` is a
        `groupby` operation or a `resample` operation (as per xarray definitions).
        Possible values: ``{"groupby", "resample", "groupby_ref_and_resample_study"}``
        (default: "resample")
        `groupby_ref_and_resample_study` may only be used when computing the
        `difference_of_means` (a.k.a the anomaly).
    indice_name: str | None
        DEPRECATED, use index_name instead.
    user_indice: dict | None
        DEPRECATED, use user_index instead.
    window_width: int
        DEPRECATED, use doy_window_width, min_spell_length or rolling_window_width
        instead.
    save_percentile: bool
        DEPRECATED, use save_thresholds instead.

    """
    _setup(callback, callback_percentage_start_value, logs_verbosity)
    (
        index_name,
        user_index,
        save_thresholds,
        doy_window_width,
    ) = _handle_deprecated_params(
        index_name,
        user_index,
        save_thresholds,
        doy_window_width,
        indice_name,
        transfer_limit_Mbytes,
        user_indice,
        save_percentile,
        window_width,
    )
    del indice_name, transfer_limit_Mbytes, user_indice, save_percentile, window_width
    # -- Choose index to compute
    interpolation = QuantileInterpolationRegistry.lookup(interpolation)
    indicator: Indicator
    standard_index: StandardIndex | None
    logical_link: LogicalLink
    coef: float | None
    build_configured_threshold = partial(
        _build_threshold,
        doy_window_width=doy_window_width,
        base_period_time_range=base_period_time_range,
        only_leap_years=only_leap_years,
        interpolation=interpolation,
    )
    if user_index is not None:
        standard_index = None
        indicator = read_indicator(user_index)
        if threshold is None:
            threshold = read_thresholds(user_index, build_configured_threshold)
        logical_link = read_logical_link(user_index)
        coef = read_coef(user_index)
        date_event = read_date_event(user_index)
        rename = index_name or user_index.get("index_name", None) or "user_index"
        output_unit = out_unit
        rolling_window_width = user_index.get("window_width", rolling_window_width)
        base_period_time_range = user_index.get(
            "ref_time_range", base_period_time_range
        )
    elif index_name is not None:
        # TODO: [BoundedThreshold] read logical_link from threshold instead
        logical_link = LogicalLinkRegistry.LOGICAL_AND
        coef = None
        standard_index = EcadIndexRegistry.lookup(index_name, no_error=True)
        if standard_index is None:
            indicator = GenericIndicatorRegistry.lookup(index_name)
            rename = None
            output_unit = out_unit
        else:
            indicator = standard_index.indicator
            threshold = standard_index.threshold
            rename = standard_index.short_name
            output_unit = out_unit or standard_index.output_unit
    else:
        raise InvalidIcclimArgumentError(
            "You must fill either index_name or user_index"
            "to compute a climate index."
        )
    sampling_frequency = FrequencyRegistry.lookup(slice_mode)
    if isinstance(threshold, str):
        threshold = build_configured_threshold(threshold)
    elif isinstance(threshold, Sequence):
        threshold = [build_configured_threshold(t) for t in threshold]
    climate_vars_dict = read_in_files(
        in_files=in_files,
        var_names=var_name,
        threshold=threshold,
        standard_index=standard_index,
    )
    # We use groupby instead of resample when there is a single variable that must be
    # compared to its reference period values.
    is_compared_to_reference = _must_add_reference_var(
        climate_vars_dict, base_period_time_range
    )
    indicator_name = (
        standard_index.short_name if standard_index is not None else indicator.name
    )
    climate_vars = build_climate_vars(
        climate_vars_dict=climate_vars_dict,
        ignore_Feb29th=ignore_Feb29th,
        time_range=time_range,
        base_period=base_period_time_range,
        standard_index=standard_index,
        is_compared_to_reference=is_compared_to_reference,
    )
    if base_period_time_range is not None:
        reference_period = tuple(
            map(lambda t: read_date(t).strftime("%m-%d-%Y"), base_period_time_range)
        )
    else:
        reference_period = None
    config = IndexConfig(
        save_thresholds=save_thresholds,
        frequency=sampling_frequency,
        climate_variables=climate_vars,
        min_spell_length=min_spell_length,
        rolling_window_width=rolling_window_width,
        out_unit=output_unit,
        netcdf_version=NetcdfVersionRegistry.lookup(netcdf_version),
        interpolation=interpolation,
        callback=callback,
        is_compared_to_reference=is_compared_to_reference,
        reference_period=reference_period,
        indicator_name=indicator_name,
        logical_link=logical_link,
        coef=coef,
        date_event=date_event,
        sampling_method=sampling_method,
    )
    result_ds = _compute_climate_index(
        climate_index=indicator,
        config=config,
        initial_history=climate_vars[0].global_metadata["history"],
        initial_source=climate_vars[0].global_metadata["source"],
        rename=rename,
        reference=standard_index.reference
        if standard_index is not None
        else ICCLIM_REFERENCE,
    )
    if out_file is not None:
        _write_output_file(
            result_ds,
            climate_vars[0].global_metadata["time_encoding"],
            config.netcdf_version,
            out_file,
        )
    callback(callback_percentage_total)
    log.ending_message(time.process_time())
    return result_ds


def _write_output_file(
    result_ds: xr.Dataset,
    input_time_encoding: dict | None,
    netcdf_version: NetcdfVersion,
    file_path: str,
) -> None:
    """Write `result_ds` to a netCDF file on `out_file` path."""
    if input_time_encoding:
        time_encoding = {
            "calendar": input_time_encoding.get("calendar"),
            UNITS_KEY: input_time_encoding.get(UNITS_KEY),
            "dtype": input_time_encoding.get("dtype"),
        }
    else:
        time_encoding = {UNITS_KEY: "days since 1850-1-1"}
    result_ds.to_netcdf(
        file_path,
        format=netcdf_version.name,
        encoding={"time": time_encoding},
    )


def _handle_deprecated_params(
    index_name,
    user_index,
    save_thresholds,
    doy_window_width,
    indice_name,
    transfer_limit_Mbytes,
    user_indice,
    save_percentile,
    window_width,
) -> tuple[str, UserIndexDict, bool, int]:
    if indice_name is not None:
        log.deprecation_warning(old="indice_name", new="index_name")
        index_name = indice_name
    if user_indice is not None:
        log.deprecation_warning(old="user_indice", new="user_index")
        user_index = user_indice
    if transfer_limit_Mbytes is not None:
        log.deprecation_warning(old="transfer_limit_Mbytes")
    if save_percentile is not None:
        log.deprecation_warning(old="save_percentile", new="save_thresholds")
        save_thresholds = save_percentile
    if window_width is not None:
        log.deprecation_warning(old="window_width", new="doy_window_width")
        doy_window_width = window_width
    return index_name, user_index, save_thresholds, doy_window_width


def _setup(callback, callback_start_value, logs_verbosity):
    # make xclim input daily check a warning instead of an error
    # TODO: it might be safer to feed a context manager which will setup
    #       and teardown these confs
    xclim.set_options(data_validation="warn")
    # keep attributes through xarray operations
    xr.set_options(keep_attrs=True)
    log.set_verbosity(logs_verbosity)
    log.start_message()
    callback(callback_start_value)


def _get_unit(output_unit: str | None, da: DataArray) -> str | None:
    da_unit = da.attrs.get(UNITS_KEY, None)
    if da_unit is None:
        if output_unit is None:
            warn(
                "No unit computed or provided for the index was found."
                " Use out_unit parameter to add one."
            )
            return ""
        else:
            return output_unit
    else:
        return da_unit


def _compute_climate_index(
    climate_index: Indicator | None,
    config: IndexConfig,
    initial_history: str | None,
    initial_source: str,
    reference: str,
    rename: str | None = None,
) -> Dataset:
    result_da = climate_index(config)
    if rename:
        result_da = result_da.rename(rename)
    else:
        result_da = result_da.rename(climate_index.name)
    result_da.attrs[UNITS_KEY] = _get_unit(config.out_unit, result_da)
    if config.frequency.post_processing is not None and "time" in result_da.dims:
        resampled_da, time_bounds = config.frequency.post_processing(result_da)
        result_ds = resampled_da.to_dataset()
        if time_bounds is not None:
            result_ds["time_bounds"] = time_bounds
            result_ds.time.attrs["bounds"] = "time_bounds"
    else:
        result_ds = result_da.to_dataset()
    if config.save_thresholds:
        result_ds = xr.merge(
            [result_ds, _format_thresholds_for_export(config.climate_variables)]
        )
    history = _build_history(result_da, config, initial_history, climate_index)
    result_ds = _add_ecad_index_metadata(
        result_ds, climate_index, history, initial_source, reference
    )
    return result_ds


def _add_ecad_index_metadata(
    result_ds: Dataset,
    computed_index: Indicator,
    history: str,
    initial_source: str,
    reference: str,
) -> Dataset:
    result_ds.attrs.update(
        dict(
            title=computed_index.standard_name,
            references=reference,
            institution="Climate impact portal (https://climate4impact.eu)",
            history=history,
            source=initial_source if initial_source is not None else "",
            Conventions="CF-1.6",
        )
    )
    result_ds.lat.encoding["_FillValue"] = None
    result_ds.lon.encoding["_FillValue"] = None
    return result_ds


def _build_history(
    result_da: DataArray,
    config: IndexConfig,
    initial_history: str | None,
    indice_computed: Indicator,
) -> str:
    if initial_history is None:
        # get xclim history
        initial_history = result_da.attrs[HISTORY_CF_KEY]
    else:
        # append xclim history
        initial_history = f"{initial_history}\n{result_da.attrs[HISTORY_CF_KEY]}"
    del result_da.attrs[HISTORY_CF_KEY]
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return (
        f"{initial_history}\n"
        f" [{current_time}]"
        f" Calculation of {indice_computed.name}"
        f" index ({config.frequency.adjective})"
        f" - icclim version: {ICCLIM_VERSION}"
    )


def _build_threshold(
    threshold: str | Threshold,
    doy_window_width: int,
    base_period_time_range: Sequence[datetime | str] | None,
    only_leap_years: bool,
    interpolation: QuantileInterpolation,
) -> Threshold:
    if isinstance(threshold, Threshold):
        return threshold
    else:
        return build_threshold(
            threshold,
            doy_window_width=doy_window_width,
            reference_period=base_period_time_range,
            only_leap_years=only_leap_years,
            interpolation=interpolation,
        )


def _format_thresholds_for_export(climate_vars: list[ClimateVariable]) -> Dataset:
    return xr.merge([_format_threshold(v) for v in climate_vars])


def _format_threshold(cf_var: ClimateVariable) -> DataArray:
    return cf_var.threshold.value.rename(cf_var.name + "_thresholds").reindex()


# TODO move `read_indicator`, `read_threshold`, `read_logical_link`, `read_coef`,
#      `read_date_event` to a user_index_parsing module


def read_indicator(user_index: UserIndexDict) -> GenericIndicator:
    calc_op = CalcOperationRegistry.lookup(user_index["calc_operation"])
    user_index_map = {
        CalcOperationRegistry.MAX: GenericIndicatorRegistry.lookup("Maximum"),
        CalcOperationRegistry.MIN: GenericIndicatorRegistry.lookup("Minimum"),
        CalcOperationRegistry.SUM: GenericIndicatorRegistry.lookup("Sum"),
        CalcOperationRegistry.MEAN: GenericIndicatorRegistry.lookup("Average"),
        CalcOperationRegistry.EVENT_COUNT: GenericIndicatorRegistry.lookup(
            "CountOccurrences"
        ),
        CalcOperationRegistry.MAX_NUMBER_OF_CONSECUTIVE_EVENTS: GenericIndicatorRegistry.lookup(  # noqa
            "MaxConsecutiveOccurrence"
        ),
        CalcOperationRegistry.ANOMALY: GenericIndicatorRegistry.lookup(
            "DifferenceOfMeans"
        ),
    }
    if calc_op == CalcOperationRegistry.RUN_SUM:
        if user_index["extreme_mode"] == "max":
            indicator = GenericIndicatorRegistry.lookup("MaxOfRollingSum")
        elif user_index["extreme_mode"] == "min":
            indicator = GenericIndicatorRegistry.lookup("MinOfRollingSum")
        else:
            raise NotImplementedError()
    elif calc_op == CalcOperationRegistry.RUN_MEAN:
        if user_index["extreme_mode"] == "max":
            indicator = GenericIndicatorRegistry.lookup("MaxOfRollingAverage")
        elif user_index["extreme_mode"] == "min":
            indicator = GenericIndicatorRegistry.lookup("MinOfRollingAverage")
        else:
            raise NotImplementedError()
    else:
        indicator = user_index_map.get(calc_op)
    if indicator is None:
        raise InvalidIcclimArgumentError(
            f"Unknown user_index calc_operation:" f" '{user_index['calc_operation']}'"
        )
    return indicator


def read_thresholds(
    user_index: UserIndexDict, _build_threshold: Callable[[str | Threshold], Threshold]
) -> Threshold | None | Sequence[Threshold]:
    thresh = user_index.get("thresh", None)
    if thresh is None or isinstance(thresh, Threshold):
        return thresh
    # todo [BoundedThreshold] re-add below code and bind to LogicalLink
    # or (
    #     isinstance(thresh, (tuple, list))
    #     and all(map(lambda th: isinstance(th, Threshold), thresh))
    # )
    logical_operation = user_index["logical_operation"]
    if not isinstance(logical_operation, (tuple, list)):
        logical_operation = [logical_operation]
    logical_operations = [OperatorRegistry.lookup(op) for op in logical_operation]
    var_type = user_index.get("var_type", None)
    if not isinstance(thresh, (tuple, list)):
        return read_threshold(thresh, var_type, logical_operations[0], _build_threshold)
    return [
        read_threshold(t, var_type, logical_operations[i], _build_threshold)
        for i, t in enumerate(thresh)
    ]


def read_threshold(
    query: str | float,
    var_type: Literal["t", "p"] | None,
    logical_operation: Operator,
    _build_threshold: Callable[[str], Threshold],
) -> Threshold:
    if isinstance(query, str) and query.endswith(PERCENTILE_THRESHOLD_STAMP):
        if var_type == USER_INDEX_TEMPERATURE_STAMP:
            replace_unit = "doy_per"
        elif var_type == USER_INDEX_PRECIPITATION_STAMP:
            replace_unit = "period_per"
        else:
            replace_unit = "period_per"  # default to period percentiles ?
        t = query.replace(PERCENTILE_THRESHOLD_STAMP, " " + replace_unit)
    else:
        t = str(query)
    return _build_threshold(str(logical_operation.operand + t))


def read_logical_link(user_index: UserIndexDict) -> LogicalLink:
    # todo add unit test using it
    logical_link = user_index.get("link_logical_operations", None)
    if logical_link is None:
        return LogicalLinkRegistry.LOGICAL_AND
    else:
        return LogicalLinkRegistry.lookup(logical_link)


def read_coef(user_index: UserIndexDict) -> float | None:
    # todo add unit test using it
    return user_index.get("coef", None)


def read_date_event(user_index: UserIndexDict) -> float | None:
    # todo add unit test using it
    return user_index.get("date_event", False)


def _must_add_reference_var(
    climate_vars_dict: dict[str, InFileDictionary],
    reference_period: Sequence[str] | None,
) -> bool:
    """True whenever the input has no threshold and only one studied variable but there
    is a reference period.
    Example case: the anomaly of tx(1960-2100) by tx(1960-1990).
    """
    t = list(climate_vars_dict.values())[0].get("thresholds", None)
    return t is None and len(climate_vars_dict) == 1 and reference_period is not None



from __future__ import annotations

from typing import Dict, List, Literal, Sequence, Tuple, Union

from xarray import DataArray, Dataset

InFileBaseType = Union[str, Sequence[str], Dataset, DataArray]
ThresholdedDict = Dict[str, Union[Dict]]  # Dict === InFileDictionary
InFileLike = Union[ThresholdedDict, InFileBaseType, Dict[str, InFileBaseType]]

FrequencyLike = Union[str, List[Union[str, Tuple, int]], Tuple[str, Union[List, Tuple]]]
# MonthsIndexer format: [12,1,2,3]
MonthsIndexer = Dict[Literal["month"], Sequence[int]]
# DatesIndexer format: ("01-25", "02-28")
DatesIndexer = Dict[Literal["date_bounds"], Tuple[str, str]]
Indexer = Union[MonthsIndexer, DatesIndexer]

SamplingMethodLike = Literal["groupby", "resample", "groupby_ref_and_resample_study"]

ThresholdValueType = Union[
    str, float, int, Dataset, DataArray, Sequence[Union[float, int, str]], None
]



from __future__ import annotations

from typing import TypedDict

from icclim.generic_indices.threshold import Threshold
from icclim.icclim_types import InFileBaseType


class InFileDictionary(TypedDict, total=False):
    """Dictionary grouping in_files and var_name functionnalities.
    It also allows to use a different input for thresholds such as percentiles.

    Examples
    --------

    .. code-block:: python

        in_files = {
            "tasmax": {
                "study": "tasmax-store.zarr",
                "threshold": build_threshold(operator=">", ["per-1.nc", "per-2.nc"]),
            },
            "pr": "pr.nc",
            "tasmin": {"study": "tasmin.nc"},
        }
    """

    study: InFileBaseType
    thresholds: Threshold | None
